import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_remove_0_order
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_mask
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_associated_legendre_matrices
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import heat2D,heat2D_multi,plot1D
from xframe import Multiprocessing
from scipy.interpolate import interp1d
from scipy.linalg import solve_triangular
from scipy.stats import unitary_group

log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
from xframe.library.mathLibrary import gsl
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_0')[0]
#analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_10p_spher')[0] 
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)
controller.chooseJob()
ccd_1 = db.load('ccd')


def init_transforms():
    cc_data = ccd_1['cross_correlation'][...,:800]
    maxQ = np.max(ccd_1['radial_points'])*2
    max_order = 99
    n_radial_points = 128#300
    opt=settings.analysis
    ht_opt={'dimensions':3,'max_order':max_order,'indices':'lm','anti_aliazing_degree':2,'n_phi':False,'n_theta':False}
    #ht_opt={'dimensions':opt.dimensions,**opt['grid'],'type':opt['fourier_transform']['type'],'pi_in_q':opt['fourier_transform']['pi_in_q'],'reciprocal_cut_off':maxQ}
    # harmonic transforms and grid
    cht=HarmonicTransform('complex',ht_opt)
    weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points,expansion_limit = 2000)
    grid_opt={'dimensions':3,'type':'Zernike_dict','max_q':maxQ,'n_radial_points':n_radial_points,**cht.grid_param,'pi_in_q':True}
    grid_pair = get_grid(grid_opt)
    maxR = grid_pair.real[:,0,0].max()
    ft,ift = generate_zernike_spherical_ft(maxR,weight_dict,cht,'complex',True,use_gpu=True)
    return ft,ift,cht,grid_pair

#locals().update(init_transforms())

#projd = db.load('reciprocal_proj_data',path_modifiers={'name':'3d_model_0','max_order':str(99)})

#p_matrices = list(projd['data_projection_matrices'])
#for order in range(projd['max_order']+1):
#    shape = p_matrices[order].shape
#    if 2*order+1 > shape[1]:
#        tmp = np.zeros((shape[0],2*order+1))
#        tmp[:,:shape[1]] = p_matrices[order]
#        p_matrices[order] = tmp 
scan_range = [1,20.1,0.5]
names = [
    '3d_pentagon_100_mO99_inverse_binned_psd20_arc_wrong_n_particles',
    '3d_pentagon_10p_spher_mO99_inverse_binned_psd20_arc_wrong_n_particles',
    '3d_pentagon_10p_spher_mO99_inverse_binned_psd20_custom_wrong_n_particles',
    '3d_model_0_mO99_cn_inverse_not_binned_psd20_custom',
    '3d_pentagon_10p_spher_mO99_inverse_binned_psd20_custom',
    '3d_pdb_2YHM_mO99_inverse_binned_psd20_custom'
]
shift = 1
n_points = 40
scales = np.logspace(np.log10(1/(10-shift+1)),np.log10(10+shift-1),n_points) #np.arange(*scan_range)
scales = np.linspace(1,20,n_points)
scales_step = scales[1]-scales[0]
radius = 250.0
name = names[0]

def scan_n_particles(scales,dataset_name,scan_range,radius,shift,**kwargs):
    mtip = controller.job.mtip
    mtip.generate_phasing_loop()
    grid_pair = mtip.grid_pair
    r = mtip.grid_pair.realGrid[:,0,0,0]
    q = mtip.grid_pair.reciprocalGrid[:,0,0,0]
    #ft,ift,cht,grid_pair = init_transforms()
    #projd = db.load('/gpfs/exfel/theory_group/user/berberic/MTIP/reciprocal_proj_data/'+dataset_name+'.h5')
    p_matrices = mtip.rprojection.projection_matrices #list(projd['data_projection_matrices'][str(i)] for i in range(projd['max_order']+1))
    #p_matrices[0]= p_matrices[0].reshape(-1,1)
    max_order = len(p_matrices)-1
    ft =mtip.process_factory.operatorDict['fourier_transform']
    cht = mtip.process_factory.operatorDict['harmonic_transform']
    icht = mtip.process_factory.operatorDict['inverse_harmonic_transform']

    spherical_formfactor = pLib.spherical_formfactor(q,radius=radius)
    neg = []
    n_pixel_in_radius = np.sum(r<=radius)
    base_guess = np.ones(grid_pair.realGrid.shape,dtype = float)
    base_guess[r>radius]=0
    I_base_guess = np.zeros_like(base_guess)
    I_base_guess[:] = spherical_formfactor[:,None,None]
    np.random.seed(int.from_bytes(os.urandom(4), byteorder='little'))
    for repeats in range(1):        
        neg_fractions = []
#        log.info('repeat count = {}'.format(repeats))
        for scale in scales:
            #log.info('creating density')
            #density_guess = np.array(base_guess)
            #density_guess[:n_pixel_in_radius]+= 0.1*np.random.rand(n_pixel_in_radius,*base_guess.shape[1:])
            #log.info('finished density of shape = {}'.format(density_guess.shape))
            #I_guess = np.abs(ft(density_guess))**2
            I_guess = np.array(I_base_guess)
            I_lm_guess = cht(I_guess)
            I_lm_guess[1:] = [ np.zeros_like(i) for i in I_lm_guess[1:]] 
            unknowns = mtip.process_factory.operatorDict['approximate_unknowns'](I_lm_guess)
            #log.info(unknowns[-1].shape)
            #I_lm = [ np.sum(p_matrices[order][:,:,None] * unknowns[order][None,:,:],axis = 1) for order in range(max_order+1)]
            I_lm = [ np.zeros_like(i) for i in I_lm_guess]
            for i in range(len(I_lm)):
                p_matrix = p_matrices[i]
                I_l = I_lm[i]
                I_l[:,:p_matrix.shape[-1]]=p_matrix
                
            I_lm[1:] = [lm/shift for lm in I_lm[1:]]
            I_lm[0]/=scale
            I = icht(I_lm)
            neg_fractions.append(np.sum(I.real<0)/np.prod(I.shape))
        neg.append(np.array(neg_fractions))
    return np.mean(neg,axis = 0)

def init_2(dataset_name):
    pd = db.load('/gpfs/exfel/theory_group/user/berberic/MTIP/reciprocal_proj_data/'+dataset_name+'.h5')
    max_order = pd['max_order']
    p_matrices = [pd['data_projection_matrices'][str(o)] for o in range(max_order +1)]
    p_matrices[0] = p_matrices[0][:,None] 
    qs = pd['data_radial_points']
    nq = len(qs)
    h_dict = {'dimensions':3,'max_order':99,'anti_aliazing_degree':2,'n_phi':False,'n_theta':False}
    cht = HarmonicTransform('complex',h_dict)
    grid = GridFactory.construct_grid('uniform',[qs,cht.grid_param['thetas'],cht.grid_param['phis']])
    integrator = mLib.SphericalIntegrator(grid[:])
    a_int = pd['average_intensity']*2*np.sqrt(np.pi)
    return locals()
#locals().update(init_2(name))

def scan_n_particles_new2():    
    I_lm = [ np.zeros((nq, 2*o+1),dtype=complex) for o in range(max_order+1)]
    for i in range(len(I_lm)):        
        I_l = I_lm[i]
        I_l[:,:p_matrices[i].shape[-1]]=p_matrices[i]

    neg_volume = np.array(tuple( integrator.integrate(cht.inverse([a_int[:,None]/s]+I_lm[1:])<0)  for s in scales))
    
    return locals()

def scan_n_particles_new2():    
    I_lm = [ np.zeros((nq, 2*o+1),dtype=complex) for o in range(max_order+1)]
    for i in range(len(I_lm)):        
        I_l = I_lm[i]
        I_l[:,:p_matrices[i].shape[-1]]=p_matrices[i]
    I = cht.inverse([a_int[:,None]]+I_lm[1:]).real
    imax = I.max()
    I/=imax
    y00=1/(2*np.sqrt(np.pi))
    a = a_int*y00/imax
    d=1
    IN = I[None,...]+(1/scales[:,None,None,None]-1)*a[None,:,None,None]
    gd1 = 2*d*IN*(a[None,:]/scales[:,None]**2)[...,None,None]
    gd2 = np.exp(-d*IN**2)
    g= gd1*gd2
    grads = [integrator.integrate(gi) for gi in g]
    #gd = 2*d*IN*(a[None,:]/scales[:,None]**2)[...,None,None]*np.exp(-d*IN**2)
    #dnIns = 2*d*Ins*a_int[None,:,None]*y00/s[:,None,None]**2*np.exp(-d*Ins**2)
    return locals()
locals().update(scan_n_particles_new2())

#neg = Multiprocessing.comm_module.request_mp_evaluation(scan_n_particles,argArrays = [scales],const_args=[names[0],scan_range,radius,shift],callWithMultipleArguments=True,n_processes = 40,splitMode='modulus')
#q_min_id=0
#neg = Multiprocessing.comm_module.request_mp_evaluation(scan_n_particles_new2,argArrays = [scales],const_args=[name,scan_range,radius,shift,q_min_id],callWithMultipleArguments=True,n_processes = 70,splitMode='modulus')


def print_neg(neg,name,radius,shift):
    expected=np.sqrt(10)
    #expected=shift
    grid = scales
    layout = {'title':'{} \n Assumed particle radius ={}'.format(name.replace('_','\_'),radius),'x_label':'scaling factor', 'y_label':'\% of negative pixels in I'}#,))*len(orders)
    neg_model = np.exp(-1/(grid)**1.12*11.3)*0.07098
    fig = plot1D.get_fig(neg,grid = grid,x_scale='lin',layout = layout)
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/n_particles/{}_rad_{}_new3.matplotlib'.format(name,radius),fig)
    layout = {'title':'{} \n Assumed particle radius ={}'.format(name.replace('_','\_'),radius),'x_label':'scaling factor', 'y_label':' diff '}#,))*len(orders)
    #diff = (neg[:-2]-neg[2:])/(scales[:-2]-scales[2:])
    acc = 8
    #diff = mLib.second_derivative(neg,scales_step,acc)
    diff = np.gradient(neg,scales_step)
    diff_model = np.gradient(neg_model,scales_step)
    fig = plot1D.get_fig(diff,grid = scales[:],x_scale='lin',layout = layout)
    ax = fig.get_axes()[0]
    ax.vlines(expected,diff.min(),diff.max())
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/n_particles/{}_rad_{}_diff_new4.matplotlib'.format(name,radius),fig)

    
    diff2 = np.gradient(diff,scales_step)
    diff2_model = np.gradient(diff_model,scales_step)
    #fig = plot1D.get_fig(diff,grid = grid[acc//2:-acc//2],x_scale='log',layout = layout)
    fig = plot1D.get_fig(diff2,grid = scales[:],x_scale='lin',layout = layout)
    ax = fig.get_axes()[0]
    ax.vlines(expected,diff.min(),diff.max())
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/n_particles/{}_rad_{}_diff2_new4.matplotlib'.format(name,radius),fig)

#print_neg(neg,name,radius,shift)



def init(): 
    thetas = ccd['data_grid']['thetas']
    phis = ccd['data_grid']['phis']
    max_order = 99
    orders = np.arange(max_order+1)
    
    metadata = {**ccd,**proj_opt,'orders':np.arange(max_order+1),'extract_odd_orders':False,'bl_enforce_psd':False}          
    cc_mask = cross_correlation_mask(ccd['data_grid'],metadata)
    #cc_mask = np.full(cc.shape,True)
    new_cc_1,new_cc_mask_1,new_phis_1 = binned_mean(np.full(cc_mask.shape,True),cc_1,max_order,phis)
    new_cc_1 = interpolate(new_cc_1,new_cc_mask_1,new_phis_1)
    cns_1 = mLib.circularHarmonicTransform_real_forward(new_cc_1)
    new_cc,new_cc_mask,new_phis = binned_mean(cc_mask,cc,max_order,phis)
    new_cc = interpolate(new_cc,new_cc_mask,new_phis)
    cns = mLib.circularHarmonicTransform_real_forward(new_cc)
    #pl,ls,ms = mLib.gsl.legendre_sphPlm_array(max_order,max_m,np.cos(thetas),return_orders = True,sorted_by_l = True)
    qq_matrix = ccd_associated_legendre_matrices(ccd['data_grid']['thetas'],max_order,max_order)
    #qq_matrix = q_matrix[None,:]*q_matrix[:,None]/(2*orders+1)[None,None,None,:]
    inverses = np.linalg.inv(qq_matrix[...,2::2,2::2])
    q_ids = np.arange(len(thetas))
    
    bls = np.zeros_like(cns)
    bls[...,2::2] = np.sum(inverses*cns[...,None,2:max_order+1:2],axis =-1)
    bls[...,1:]/=metadata['n_particles']
    bls[...,0]/=metadata['n_particles']**2
    #bls_pseudo = np.swapaxes(cross_correlation_to_deg2_invariant(cc,3,**metadata)[0],0,-1)
    #Ms = qq_matrix[...,2::2,2::2]
    #Cs = cns[...,2:max_order+1:2]
    #bls[...,2::2] = np.array(tuple(solve_triangular(m,v) for m,v in zip(Ms.reshape(-1,*Ms.shape[-2:]),Cs.reshape(-1,*Cs.shape[-1:])) )).reshape(*Cs.shape)
    #bls_t = np.array(tuple(mLib.optimal_tikhonov_regularization(qq_matrix[i],cns[i],allow_offset=True)[0] for i in range(len(bls))))
    #bls_t = Multiprocessing.comm_module.request_mp_evaluation(mLib.optimal_tikhonov_regularization_worker,argArrays = [q_ids,q_ids],const_args=[qq_matrix,cns[...,:max_order+1].astype(complex),True],callWithMultipleArguments=True)
    cns_new = np.zeros_like(cns)
    #cns_new[...,2::2] = np.sum(qq_matrix[...,2::2,2::2]*bls[...,None,2:max_order+1:2],axis =-1)
    #cns/=10
    return locals()

#locals().update(init())

def plot():
    qs = ccd['data_grid']['qs']
    qq_grid = GridFactory.construct_grid('uniform',[qs,qs])
    orders = np.arange(2,12,2)
    orders2 = np.arange(12,22,2)
    bls_plot = np.swapaxes(bls,0,-1)#/100**2
    bls_plot[0]#/=100
    cns_plot = np.swapaxes(cns,0,-1)#/100**2
    cns_plot_1 = np.swapaxes(cns_1,0,-1)#/100**2
    cns_new_plot = np.swapaxes(cns_new,0,-1)#/100**2
    cns_diff = np.abs(cns_plot - cns_new_plot)/np.abs(cns_plot)
    cns_diff_1_100 = np.abs(cns_plot_1-cns_plot)/np.abs(cns_plot_1)
    bls_plot_psd = np.zeros_like(bls_plot)
    q_start = 25
    bls_plot_psd[:,q_start:,q_start:] = mLib.nearest_positive_semidefinite_matrix(bls_plot[:,q_start:,q_start:])
    layout = {'title':'','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':20}#,))*len(orders)
    #bls_plot[:,30:,30:] = mLib.nearest_positive_semidefinite_matrix(bls_plot[:,30:,30:])
    fig = heat2D_multi.get_fig((
        list(np.abs(cns_plot_1[orders])),
        list(np.abs(bls_plot[orders])),
        list(np.abs(bls_plot_psd[orders]))
    ),grid = qq_grid,shape = (3,len(orders)),size=(5*len(orders)+2*5,3*5),vmin = 1e+14, vmax = 1e+24,scale='log',cmap='plasma',layout = layout)
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl0-10_10p_mO70_rect.matplotlib',fig)
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl0-20_10p_mO99_rect.matplotlib',fig)
    #fig = heat2D_multi.get_fig((
    #    list(np.abs(cns_diff_1_100[orders])[:,1:,1:])
    #    ,),grid = qq_grid[1:,1:],shape = (1,len(orders)),size=(5*len(orders)+2*5,1*5),cmap='plasma')
    fig = heat2D_multi.get_fig((list(np.abs(cns_diff_1_100[orders])),list(np.abs(cns_diff_1_100[orders2]))),grid = qq_grid,shape = (2,len(orders)),size=(5*len(orders)+2*5,2*5),vmin = 1e-2, vmax = 1e2,scale='log',cmap='plasma',layout = layout)
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/cn0-10_p1_p10_mO30_diff_rect.matplotlib',fig)
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/cn0-20_p1_p10_mO99_diff_spher.matplotlib',fig)
    #fig = heat2D_multi.get_fig((
    #    list(np.abs(cns_diff[orders]))
    #),grid = qq_grid,shape = (1,len(orders)),size=(5*len(orders)+2*5,1*5),scale='log',cmap='plasma')    

#plot()



import numpy as np
import time
from xframe.settings import general
general.n_control_worker=0
import xframe
from os import path as pa

db = xframe.database.default

path = '/gpfs/exfel/theory_group/user/berberic/cryst_vols/sphv_agno3-d07_h_final.npy'
_dir = pa.dirname(path)
data = np.load(path,allow_pickle = True)

n_qs = 300
qs = np.arange(n_qs)*9/n_qs
n_phi = 500
phi=np.arange(n_phi)/n_phi *2*np.pi
n_theta = 1000
phi=np.arange(n_thetas)/n_thetas *np.pi
grid = xframe.lib.grid.GridFactory.construct_grid('uniform',[qs,thetas,phis])[:]

db.save(_dir+'volume.vts',[data.reshape(len(qs,n_theta,n_theta))],grid_type='sperical')


xframe.setup_experiment('fxs3046_online','normalized_proc')








import os
import sys
from importlib import util
from . import general
from .tools import DictNamespace

project = DictNamespace.dict_to_dictnamespace({}) 
_project_at_run = False
_project_at_select = False
raw_project = DictNamespace.dict_to_dictnamespace({})

experiment = DictNamespace.dict_to_dictnamespace({}) 
_experiment_at_run = False
_experiment_at_select = False
raw_experiment = DictNamespace.dict_to_dictnamespace({}) 

# only to make it the same type as the other settings #
gd = {key:general.__getattribute__(key) for key in dict(general.__dict__).keys() if key[0]!="_"}
general = DictNamespace.dict_to_dictnamespace(gd)

def _update_settings_on_project_select(opt,raw_opt):
    global project
    global raw_project
    global _project_at_run
    global _project_at_select
    project = opt
    raw_project = raw_opt
    _project_at_run = False
    _project_at_select = opt.dict()
    
def _update_settings_on_experiment_select(opt,raw_opt):
    global experiment
    global raw_experiment
    global _experiment_at_run
    global _experiment_at_select
    experiment = opt
    raw_experiment = raw_opt
    _experiment_at_run = False
    _experiment_at_select = opt.dict()

def _save_settings_on_controller_run():
    global _project_at_run 
    _project_at_run = project.dict()
    global _experiment_at_run
    _experiment_at_run = experiment.dict()

def get_settings_to_save():    
    project_save=raw_project
    if isinstance(_project_at_select,dict) and isinstance(_project_at_run,dict):
        project_settings_where_modified = _project_at_select != _project_at_run
        if project_settings_where_modified:
            project_save = _project_at_run
    experiment_save=raw_experiment
    if isinstance(_experiment_at_select,dict) and isinstance(_experiment_at_run,dict):
        experiment_settings_where_modified = _experiment_at_select != _experiment_at_run
        if experiment_settings_where_modified:
            experiment_save = _experiment_at_run
    return project_save,experiment_save



import numpy as np
import sys
import os
from os import path as op
import logging
import traceback
#import glob
import struct
import re
import glob
import shutil
from importlib import util
pjoin = os.path.join
pexists = os.path.exists

from xframe.settings import general as xframe_opt

# Custom data types that may need to be saved #
from xframe.library.gridLibrary import NestedArray
from xframe.library.pythonLibrary import FTGridPair,DictNamespace

# Interfaces that the default DB should satisfy #
from xframe.interfaces import DatabaseInterface

# Interfaces that will be replaced upon depencency injection of the external plugins that satisfy them. #
from xframe.database.interfaces import VTKDependency
from xframe.database.interfaces import PDBDependency
from xframe.database.interfaces import HDF5Dependency
from xframe.database.interfaces import YAMLDependency
from xframe.database.interfaces import OpenCVDependency
from xframe import Multiprocessing 

# dependency injection attributes #
# Their correct implementations will be inserted by xframe/__init__.py on runtime#
VTK_saver = VTKDependency
PDB_loader = PDBDependency
HDF5_access = HDF5Dependency
YAML_access = YAMLDependency
OpenCV_access = OpenCVDependency

log=logging.getLogger('root')

class FileAccess():
    '''
    Class that handles folders and file paths that will be used to access data.
    It's main function is to read in folder and file paths form the settings file
    and provide access to them via the 'get_path' method
    '''
    def __init__(self,folders={},files={}):
        self.folders=self.parse_folders(folders)
        self.files=files

    def set_folders_and_files(self,folders={},files={}):
        self.folders=self.parse_folders(folders)
        self.files=files
    def update_folders_and_files(self,folders={},files={}):
        self.folders.update(self.parse_folders(folders))
        self.files.update(files)
    def get_path(self,name,path_modifiers={},is_file=True,**kwargs):     
        try:
            if is_file:
                #log.info('requested path name = {}'.format(name))
                _file=self.files[name]
                location=self.folders[_file['folder']]
                file_name=_file['name']
                #log.info('location = {}'.format(location))
                #log.info('file_name = {}'.format(file_name))
                path=location+file_name
            else:
                path=self.folders[name]

            path=self.modify_path(path,modifiers=path_modifiers)
            path = os.path.expanduser(path)
        except KeyError as e:
            #traceback.print_exc()
            #log.info()
            #log.error(e)
            log.debug('No path specified in settings for name "{}".Try to use it as direct path'.format(name))
            path = None
            if isinstance(name,str):
                if len(name)>0:
                    path = os.path.expanduser(name)
            if path is None:
                raise AssertionError(f'Saving to path {name} is not possible. Wrong path definition.')           
        return path
    def create_path_if_nonexistent(self,path):
        dirname=os.path.dirname(path)
        if not (os.path.exists(dirname) or dirname==''):
            try:
                os.makedirs(dirname)
            except FileExistsError:
                pass
    def modify_path(self,path,modifiers={}):
        try:
             path=path.format(**modifiers)
        except KeyError as e:
            log.error(e)
#            n_modifiers = min( (path.count('{'),path.count('}')) )
            log.error('given file path: {} contains the modifiers {}, some of them are wrong.'.format(path,modifiers))
            raise e
        return path

    def parse_folders(self,folder_dict):
        known_folders={}
        for i in range(len(folder_dict)):
            for key,item in folder_dict.items():
                if isinstance(item,str):
                    known_folders[key]=item
                elif isinstance(item,(dict,DictNamespace)):
                    path = self.recursive_folders(folder_dict[key],known_folders)
                    if isinstance(path,str):
                        known_folders[key]=path
        return known_folders
    def recursive_folders(self,folder_dict,known_folders):
        for key in folder_dict:
            if key in known_folders:            
                path=known_folders[key]
                if isinstance(folder_dict[key],(dict,DictNamespace)):
                    path_result=self.recursive_folders(folder_dict[key],known_folders)                
                    if isinstance(path_result,str):                    
                        path+=path_result
                    else:
                        path=False
                elif isinstance(folder_dict[key],str):
                    path+=folder_dict[key]
            else:
                path = False
        return path

    def create_symlink(self,data_path,sym_path,is_file = True,path_modifiers = {}):
        data_path = self.get_path(data_path,is_file=is_file,path_modifiers = path_modifiers)
        sym_path = self.get_path(sym_path,is_file=is_file,path_modifiers = path_modifiers)
        data_path_exists = os.path.exists(data_path)
        sym_path_exists = os.path.exists(sym_path)
        #log.info('data exists = {} path = {}'.format(data_path_exists,data_path))
        #log.info('sym path = {}'.format(sym_path))
        if not sym_path_exists:
            self.create_path_if_nonexistent(sym_path)
        if data_path_exists and (('/' in sym_path) or ('\\' in sym_path)):
            if sym_path_exists:
                os.remove(sym_path)
            os.symlink(data_path,sym_path)

    def copy(self,source,destination,path_modifiers = {}):
        source = self.get_path(source,path_modifiers=path_modifiers)
        destination = self.get_path(destination,path_modifiers=path_modifiers)
        shutil.copyfile(source,destination)
        
    def create_folders(self,path,path_modifiers={}):
        path = self.get_path(path,is_file=False,path_modifiers = path_modifiers)
        self.create_path_if_nonexistent(path)
class DefaultDB(FileAccess,DatabaseInterface):
    '''
    Modular loading and saving Class which can be inherited from to generate customiced loading and saving routines.
    Main methods are 'save' and 'load'.
    
    If load or save are called with the argument <name>(and sone data for save) a 4 step process starts.
    1. Try to find a method in this instance which is called load_<name>, save_<name> and execute it.
    2. else lookup if there is a stored data path to the given <name> if yes retreive that path
    3. else assume <name> itself is a complete file path.
    4. Check if there is an available plugin for the given filetype, if yes call it.

    If one inherits from default_DB point 1. in above sequence lets you easily do some pre or postprocessing if the data
    you want to retrieve is not completely in the shape you want it to be.
    
    A new access method can be added to the access_methods attribute by specifiing the Abstract class and the filetype which it is supposed to handle.

    Note that no direct dependencies on external libraries should never be placed here. Use dependency injection for that.
    See xframe/__init__.py , xframe/externalLibraries/ and xframe/database/interfaces.py
    '''
    
    def __init__(self,**folders_files):
        super().__init__(**folders_files)
        self.access_method_splitter = '://'
        self.data_type_splitter = '.'
        self.default_access_method_str = 'file'
        self.default_data_type_str = 'default'
        self.access_methods={
            'file://':{
                'h5':'HDF5_access',
                'vtk':'VTK_saver',
                'vts':'VTK_saver',
                'vtr':'VTK_saver',
                'npy':'NumpyAccess',
                'raw':'BinaryAccess',
                'txt':'TextAccess',
                'bash':'TextAccess',
                'sh':'TextAccess',
                'zsh':'TextAccess',
                'fish':'TextAccess',
                'yaml':'YAML_access',
                'matplotlib':'MatplotlibAccess',
                'cv':'OpenCV_access',
                'py':'PythonAccess',
                self.default_data_type_str:'TextAccess'
            },
            'pdb://':{
                self.default_data_type_str:'PDB_loader'
            }
        }    
    def load(self,name,**kwargs):        
        try:
            loader = self.get_method(name,prefix = 'load_')
            path = name
            file_options=self.files.get(path,{}).get('options',{})
            if isinstance(loader,bool) or kwargs.get('skip_custom_methods',False):
                #log.info(name)
                path = self.get_path(name,**kwargs)
                #log.info(path)
                if isinstance(path,bool):
                    path = name
                loader = self.get_db(path).load
                path = self.remove_access_method_from_path(path)
            #log.debug(loader)
            #log.debug(path)
            data = loader(path,**{**kwargs,**file_options})
        except Exception as e:
            if not isinstance(e,FileNotFoundError):
                traceback.print_exc()
            if not (path in locals()):
                path = "not defined. Error during path definition."
            if not (loader in locals()):
                loader = "not defined. Could not find loader."            
            log.info('Failed to load for name: {} \n path: {}\n loader: {}\n with error: {}'.format(name,path,loader,e))
            raise e
        return data
        
    def save(self,name,data,**kwargs):
        try:            
            saver = self.get_method(name,prefix = 'save_')
            path = name
            file_options=self.files.get(path,{}).get('options',{})
            if isinstance(saver,bool) or kwargs.get('skip_custom_methods',False):
                path = self.get_path(name,**kwargs)
                if isinstance(path,bool):
                    path = name
                saver = self.get_db(path).save
                path = self.remove_access_method_from_path(path)
                self.create_path_if_nonexistent(path)
            log.debug('saving "{}" to path:  {}'.format(name,path))
            return_value = saver(path,data,**{**kwargs,**file_options})
            return return_value
        except Exception as e:
            traceback.print_exc()
            log.error(e)
            log.error('saving for "{}" failed.'.format(name))

    def load_direct(self,name,**kwargs):
        return self.load(name,skip_custom_methods=True,**kwargs)
    def save_direct(self,name,data={},**kwargs):
        return self.save(name,data,skip_custom_methods=True,**kwargs)

        
    def remove_access_method_from_path(self,path):
        access_method_split = path.split(self.access_method_splitter)
        if len(access_method_split)>1:
            new_path = access_method_split[-1]
        else:
            new_path = path
        return new_path
        
    def get_db(self,path):
        a_split = self.access_method_splitter
        access_method_split = path.split(a_split)
        if len(access_method_split)>1:
            access_method = access_method_split[0]+a_split
        else:
            access_method = self.default_access_method_str+a_split

        d_split = self.data_type_splitter
        name = os.path.basename(path)
        if name[0]==self.data_type_splitter:
            name=name[1:]
        data_type_split = name.split(d_split)
        if len(data_type_split)>1:
            data_type = data_type_split[-1]
        else:
            data_type = self.default_data_type_str            
            
        db = globals()[self.access_methods[access_method][data_type]]
        #log.info(access_method_split)
        #log.info(data_type_split)
        #log.info('{},{}'.format(access_method,data_type))
        #log.info(db)
        return db
    
    def get_method(self,name,prefix=''):
        loader = False
        try:
            loader=getattr(self,prefix+name)
        except AttributeError as e:
            pass
            #log.info('No db method found for "{}". Use default.'.format(prefix+name))
        return loader

    def recursive_command_execution(self,settings_dict,parent_dict,parent_key):
        for key in settings_dict:
            if key == 'command':
                parent_dict[parent_key]=eval(settings_dict[key])
            elif isinstance(settings_dict[key],dict):
                self.recursive_command_execution(settings_dict[key],settings_dict,key)

    def assert_path_exists(self,path):
        if not os.path.exists(path):
            raise FileNotFoundError('The file "{}" does not exist.'.format(path))
        
    def load_settings(self,name,project_path='',worker_name='',settings_file_name=False,target='project',direct_path = False,ignore_file_not_found=False):
        loader = globals()[self.access_methods['file://']['yaml']].load
        dir_name = os.path.basename(os.path.dirname(project_path))
        settings_folders = [
            self.get_path('settings_'+target,path_modifiers={target:dir_name,'worker':worker_name},is_file=False),
            self.get_path('settings_direct',path_modifiers={'path':project_path,'worker':worker_name},is_file=False),
            self.get_path('settings_default_'+target,path_modifiers={target:dir_name,'worker':worker_name},is_file=False),
            self.get_path('settings_install_'+target,path_modifiers={target:dir_name,'worker':worker_name},is_file=False),      
        ]
        s_load = SettingsLoader(loader,settings_folders,ignore_file_not_found=ignore_file_not_found)
        settings,defaults,loader_sub_defaults = s_load.load_settings_files(project_path,worker_name,settings_file_name,direct_path = direct_path)
        s_parser = SettingsParser(loader_sub_defaults)
        settings_out,raw_out = s_parser.parse(settings,defaults)
        return settings_out,raw_out

    def format_settings(self,settings):
        return YAML_access.format_settings_dict(settings)
    def update_settings(self,opt):
        if isinstance(opt,DictNamespace):
            opt = opt.dict()
        self.update_folders_and_files(**opt.get('IO',{}))


# Simple acces Access classes #
class MatplotlibAccess:
    @staticmethod
    def save(path,data,dpi=300,bbox_inches='tight',_format='png',**kwargs):
        path_with_correct_filetype = op.splitext(path)[0]+'.'+_format
        data.savefig(path_with_correct_filetype,dpi=dpi,format=_format,bbox_inches=bbox_inches)
    @staticmethod
    def load(path,**kwargs):
        e = NotImplementedError('loading with the Matplotlib tool is not possible.')
        raise e        

class NumpyAccess:
    @staticmethod
    def save(path,data,**kwargs):
        np.save(path,data)
    @staticmethod
    def load(path,**kwargs):
        return np.load(path)
class BinaryAccess:
    @staticmethod
    def save(path,data,**kwargs):
        endian = kwargs.get('endian','<')            
        fmt=endian+str(data.size)+data.dtype.char            
        bin_struct = struct.pack(fmt, *data.flatten()) 
        with open( path, "wb") as file:
            file.write( bin_struct )
    @staticmethod
    def load(path,**kwargs):
        with open( path, "rb") as file:
            data_bin = file.read()
        if len(data_bin)!=0:
            dtype = np.dtype(kwargs.get('dtype',np.float32))
            endian = kwargs.get('endian','<')            
            length = int(len(data_bin)//dtype.itemsize)
            fmt=endian+str(length)+dtype.char
            data=np.asarray(struct.unpack(fmt, data_bin),dtype = dtype)
        else:
            data = numpy([])
        return data

class TextAccess:
    @staticmethod
    def save(path,data,**kwargs):
        with open(path,'wt') as txt_file:
            if isinstance(data,str):
                txt_file.write(data)
            elif isinstance(data,(list,tuple)):
                txt_file.writelines(data)
            
    @staticmethod
    def load(path,**kwargs):
        with open(path,'rt') as txt_file:
            text = txt_file.readlines()
        return text
class PythonAccess:
    @staticmethod
    def load(path,as_text=False,**kwargs):
        if as_text:
            module = TextAccess.load(path,**kwargs)
        else:
            spec = util.spec_from_file_location('file',general.settings_file)
            module = util.module_from_spec(spec)
            spec.loader.exec_module(module)
        return module    
    @staticmethod
    def save(path,data,**kwargs):
        if isinstance(data,(str,list,tuple)):
            TextAccess.save(path,data,**kwargs)
        else:
            raise NotImplementedError('Saving of python modules as .py files is not supported yet.')

    
# Settings management #
class SettingsLoader:
    def __init__(self,file_loader,settings_folders,ignore_file_not_found=False):
        self.load_file = file_loader
        self.settings_folders = settings_folders
        self.ignore_file_not_found = ignore_file_not_found
        
    def get_settings_path(self,folders,settings_file_name):
        settings_paths = [pjoin(f+settings_file_name+'.yaml') for f in folders]
        settings_path = False
        for folder in folders:
            path = pjoin(folder,settings_file_name+'.yaml')
            if pexists(path):
                settings_path = path                
                break
        
        if isinstance(settings_path,bool):
            message = f'Non of the possible possible settings files exist! They are {settings_paths}'
            if self.ignore_file_not_found:
                message+='Try to start with unchanged default settings.'
            else:
                raise FileNotFoundError(message)
        return settings_path,folder
    
    def get_default_settings_path(self,folder,input_version=False,name = '',warn = False):
        regexpr = xframe_opt.default_settings_regexpr
        if len(name)>0:
            search_expression = pjoin(folder,'default_'+name+'_'+'[0-9]*.yaml')
        else:
            search_expression = pjoin(folder,'default_[0-9]*.yaml')
        default_paths = sorted(glob.glob(search_expression))
        version_number_dict = {re.search(regexpr,p):p for p in default_paths}
        if len(default_paths)>0:
            latest_path = default_paths[-1]
            default_path = latest_path
            if not isinstance(input_version,bool):
                if str(input_version) in version_number_dict:
                    default_path = version_number_dict[str(input_version)]
                elif warn:
                    log.warning(f'Default settings for specified settings_version {input_version} not found. Try to use latest known default file at: {latest_path}. Worker name {worker_name}')
        else:
            if warn:
                log.info(f'No defaults found at {search_expression} . Trying to continue with empty default.')
            default_path = False
        return default_path

    def generate_load_sub_defaults(self,settings_version):
        folder = str(self.selected_folder)        
        def load(name):
            path = self.get_default_settings_path(folder,input_version=settings_version,name = name)
            if isinstance(path,str):
                return self.load(path)
            else: 
                return {}
        return load

    def load_settings_files(self,path,worker_name,settings_file_name,target='project',direct_path = False,fail_silent=False):
        if not isinstance(direct_path,str):
            settings_folders = self.settings_folders
            if isinstance(settings_file_name,str):
                settings_path,folder_path = self.get_settings_path(settings_folders,settings_file_name)                
            else:
                settings_path = False
                folder_path=self.settings_folders[-1]
            self.selected_folder=folder_path            
            if isinstance(settings_path,str):
                settings = self.load_file(settings_path)
            else:
                settings = {}
            settings_version = settings.get(xframe_opt.settings_version_key,False)
            default_path = self.get_default_settings_path(settings_folders[0],input_version=settings_version,warn = True)
            if isinstance(default_path,str):            
                default_settings = self.load_file(default_path)
            else:
                default_settings = {}
            self.default_settings_path = default_path
            self.settings_path = settings_path
            if not isinstance(default_path,bool):
                default_version = re.search(xframe_opt.default_settings_regexpr,os.path.basename(default_path)).group()
            else:
                default_version = False
        else:
            settings = self.load_file(direct_path)
            self.selected_folder=os.path.dirname(direct_path)            
            default_settings = {}
            default_version = False
            settings_path = direct_path
            default_path = ''
        load_sub_defaults = self.generate_load_sub_defaults(default_version)
        settings['_settings_path']=settings_path
        settings['_default_settings_path']=default_path
        return settings,default_settings,load_sub_defaults

class SettingsParser:
    def __init__(self,load):
        self.load = load
        self.parse_order = ['_only_if','_copy','_if']

    def recursive_command_execution(self,settings_dict,parent_dict,parent_key):
        for key in settings_dict:
            if key == 'command':
                #log.info(settings_dict[key])
                parent_dict[parent_key]=eval(settings_dict[key])
            elif isinstance(settings_dict[key],dict):
                self.recursive_command_execution(settings_dict[key],settings_dict,key)

    
    def parse_path(self,current_path,relative_path):
        '''
        processes relative path definition at current place in defaults file.
        '''
        if relative_path[0]!="/":
            if len(current_path) >0:
                tmp_path = relative_path.split('../')
                backwards_index = -len(tmp_path)+1
                if backwards_index ==0:
                    backwards_index = None
                path = current_path[:backwards_index]+tmp_path[-1].split('/')
            else:
                path = [relative_path]
        else:
            path = relative_path[1:].split('/')
        return path

    def get_value(self,dataset,current_path,relative_path):
        '''
        get value in out at position given as a relative path.
        '''
        keys = self.parse_path(current_path,relative_path)
        #print('get path dataset: ',dataset)
        #print('get path current: ',current_path)
        #print('get path keys: ', keys)
        value = dataset[keys[0]]
        for key in keys[1:]:
            value = value[key]
        return value
    
    def set_dict_entry(self,dataset,value,current_path,relative_path):
        '''
        get value in out at position given as a relative path.
        '''
        keys = self.parse_path(current_path,relative_path)
        #print('get path dataset: ',dataset)
        #print('get path current: ',current_path)
        #print('get path keys: ', keys)
        sub_dict = dataset[keys[0]]
        for key in keys[1:-1]:
            sub_dict = sub_dict[key]
        sub_dict[keys[-1]]=value
        
    def get_default_value(self,out,current_path,sub_default):
        '''
        In the default settings file get value for an entry that has no sub settings.
        '''
        default_value = sub_default['_value']
        #log.info(f'default value = {default_value}')
        if isinstance(default_value,(dict,DictNamespace)):
            if '_copy' in default_value:
                opt = default_value['_copy']
                relative_path = opt
                value=self.get_value(out,current_path,relative_path)
            #mode,opt = list(default_value.items())[0] 
            #if mode == '_copy':
            #    relative_path = opt
            #    value=self.get_value(out,current_path,relative_path)
            else:
                return default_value
                #log.error(f'Wrong format in default entry {current_path+[sub_default.dict()]} under _value only "command:" entry is allowed as sub directory but it is {default_value.dict()}')
        else:
            value = default_value
        return value

                
    def check_values(self,*args,**kwargs):
        '''
        future option to check data types against specification in defaults file.
        '''
        return True
    

    ## dynamic default modifiers ##
    def _if(self,out,default,sub_default,current_path):
        value_path=sub_default['_if']['x']
        #log.info(f' current_path = {current_path[:-1]}, \n value_path = {value_path} ')
        x = self.get_value(out,current_path[:-1],value_path)
        #log.info(f'value in if is x = " {x} "')
        conditions = sub_default['_if']['condition']
        if not isinstance(conditions,list):
            conditions = [conditions]
        true_index = len(conditions)
        for _id,c in enumerate(conditions):
            if eval(c):
                true_index = _id
                break
        #print('true index = ',true_index)
        value = sub_default['_if']['values'][true_index]
        
        return {"_value":value},False
        
    def _only_if(self,out,default,sub_default,current_path):
        value_path=sub_default['_only_if']['x']
        x = self.get_value(out,current_path[:-1],value_path)
        condition = sub_default['_only_if']['condition']
        skip = not bool(eval(condition))
        return sub_default,skip
    
    def _copy(self,out,default,sub_default,current_path):
        value_path = sub_default.pop('_copy')
        #print('copy value path: ',value_path)
        try:
            value = self.get_value(default,current_path[:-1],value_path)
        except Exception as e:
            log.error(f'current path {current_path}, value path = {value_path}, \n sub_default = {sub_default},\n default = {default}')
            raise e
        self.set_dict_entry(default,value,current_path,f'../{current_path[-1]}')
        #log.info(f'copyed part: \n \n {value}')
        #d,skip = self.parse_default_sketch(out,default,value,current_path)
        #log.info(f'processed copyed part: \n \n {d}')
        return value,False


    ## parse defaults ##
    def parse_default_sketch(self,out,default,sub_default_sketch,current_path):
        skip = False
        for method_key in self.parse_order:
            try:
                #log.info(f'method key: {method_key}')
                #log.info(f'sketch: {sub_default_sketch}')
                if method_key in sub_default_sketch:
                    #log.info(f'found: {method_key}')
                    try:
                        method = self.__getattribute__(method_key)
                    except AttributeError:
                        continue
                    sub_default_sketch,skip = method(out,default,sub_default_sketch,current_path)
                    if skip:
                        break
            except Exception as e:
                log.error(f'error in step: {method_key}, error: {e}, Traceback:\n')
                traceback.print_exc()
        return sub_default_sketch,skip
    def apply_defaults(self,data,current_path=False,sub_data = False):
        if (not isinstance(current_path,list)) or (not isinstance(sub_data,list)):
            sub_data=data
            current_path = []
            
        defaults,settings,out = sub_data
        global_out = data[-1]
        global_default = data[0]
        #print('def: ' ,defaults)
        if '_import' in defaults:
            import_name = defaults.pop('_import')
            imported_defaults = DictNamespace.dict_to_dictnamespace(self.load(import_name))
            defaults.update(import_defaults)
        if not isinstance(defaults,(dict,DictNamespace)):
            log.info(defaults)
        for key,d in defaults.items():
            try:
                d,skip = self.parse_default_sketch(global_out,global_default,d,current_path+[key])
                if skip or (key[0]=='_'):
                    continue
                exists = key in settings
                d_contains_sub_settings = ('_value' not in d)
                if exists:
                    value = settings[key]
                    s_contains_sub_settings = isinstance(value,(dict,DictNamespace))
                    if s_contains_sub_settings and d_contains_sub_settings:
                        self.apply_defaults(data,current_path = current_path + [key],sub_data = [d,settings[key],out[key]])
                    elif (not s_contains_sub_settings) and (not d_contains_sub_settings):
                        value_valid = self.check_values(d,value)
                        if not value_valid:
                            out[key]=d['_value']
                else:
                    if d_contains_sub_settings:
                        out[key]={}
                        self.apply_defaults(data,current_path = current_path + [key],sub_data = [d,{},out[key]])
                    else:
                        out[key] = self.get_default_value(global_out,current_path,d)
            except Exception as e:
                log.error(f'Error in Application of default settings at in path {current_path} at key {key}')
                traceback.print_exc()
                raise e
        return out
    
    def parse(self,settings_dict,default_settings_dict):
        raw_settings = DictNamespace.dict_to_dictnamespace(settings_dict)
        raw_defaults = DictNamespace.dict_to_dictnamespace(default_settings_dict)
        self.recursive_command_execution(settings_dict,{'value':0},'value')
        self.recursive_command_execution(default_settings_dict,{'value':0},'value')
        settings = DictNamespace.dict_to_dictnamespace(settings_dict)
        default_settings = DictNamespace.dict_to_dictnamespace(default_settings_dict)
        settings_out = settings.copy()
        settings_out = self.apply_defaults([default_settings,settings,settings_out])
        raw_out = raw_settings.copy() 
        raw_out = self.apply_defaults([raw_defaults,raw_settings,raw_out])
        return settings_out,raw_out
            



from xframe.interfaces import DependencyMeta
import abc

#class DependencyInterfaceTemplate:
#    pass

class MatplotlibInterface(abc.ABC):
    _external_dependency_name_ = 'matplotlib'

class MatplotlibDependency(MatplotlibInterface,metaclass = DependencyMeta):
  pass
  

class MPLToolkitInterface(abc.ABC):
    _external_dependency_name_ = 'matplotlib_toolkit'
class MPLToolkitDependency(MPLToolkitInterface,metaclass = DependencyMeta):
  pass
class OpenCVInterface(abc.ABC):
    _external_dependency_name_ = 'opencv'
    __dependency__='cv2 (opencv-python)'
    @abc.abstractmethod
    def get_polar_image(*args,**kwargs):
        pass
class OpenCVDependency(OpenCVInterface,metaclass = DependencyMeta):
  pass



import logging
import numpy as np

from .interfaces import DetectorInterfaceSimulation
from xframe.experiment.interfaces import SimulationInterface

log=logging.getLogger('root')



class RandomSimulator(SimulationInterface):
    def __init__(self,detector,experimentalSetup):
        try:
            assert isinstance(detector,DetectorInterfaceSimulation)
        except AssertionError:
            log.error('The Detector {0} given to {1} does not implement {2}.'.format(detector,type(self).__name__,DetectorInterfaceSimulation.__name__))
            raise
        self.detector=detector

        try:
            assert 'SampleDetectorDistance' in experimentalSetup.keys()
        except AssertionError:
            log.error('experimentalSetup does not contain the Sample Distance')
            raise
        self.experimentalSetup=experimentalSetup
        
    def getData(self,*args,**kwargs):
        amplitude=kwargs.get('Amplitude',1)
        pixelArrayShape=self.detector.getGeometry()['PixelKornerIndex'].shape
        randomImage=amplitude*np.random.rand(pixelArrayShape[0],pixelArrayShape[1])
        return randomImage

    def getGeometry(self):
        geometry=self.detector.getGeometry()
        return geometry

    def getExperimentalSetup(self):
        return self.experimentalSetup


class MockSimulator(SimulationInterface):
    def getData(self):
        pass
    def getGeometry(self):
        pass
    def getExperimentalSetup(self):
        pass



if os.getcwd()[-6:] != "source":
    #print(os.getcwd())
    os.chdir('../')
    
import numpy as np
from analysisLibrary import saxs as saxs_lib


data = np.arange(100).reshape(10,10)
mask = (np.random.rand(100).reshape(10,10)+0.5).astype(int).astype(bool)
mask[0] = False

saxs = saxs_lib.calc_saxs(data,mask)
saxs_slow = saxs_lib.calc_saxs_slow(data,mask)

print( 'both saxs versions give the same answer ={}'.format((saxs == saxs_slow).all())) 


data=np.array([[0,1,0,1,1,0]])
mask= data.astype(bool)
saxs = saxs_lib.calc_saxs(data,mask)
expected = np.array([1]) 
print('saxs is as expected = {}'.format((saxs == expected).all()))



import time
import numpy as np
from scipy.interpolate import griddata
import logging
import traceback

log=logging.getLogger('root')

from xframe.library.physicsLibrary import ewald_sphere_theta
from xframe.library.gridLibrary import GridFactory
from xframe.library.mathLibrary import cartesian_to_spherical
from xframe.library.mathLibrary import spherical_to_cartesian
from xframe.library import units

def find_center_by_point_sym_overlap(data,grid,mask = True):
    log.info('\n mask_data dtype 3 ={}'.format(mask.dtype))
    log.info('input nonzero = {}'.format((data!=0).any()))
    n_x_points = grid.shape[0]
    n_y_points = grid.shape[1]
    zero_index =np.array(np.nonzero((np.isclose(grid[...,0],0))&(np.isclose(grid[...,1],0)))).flatten()
    shift = -zero_index +[n_x_points//2,n_y_points//2]

    data = np.roll(np.roll(data,shift[0],axis = 0),shift[1],axis=1) # correct off center origin for point inversion
    data = np.pad(data,((n_x_points,0),(n_y_points,0)))
    r_data = np.fft.fft2(data)
    convolution =  np.fft.ifft2(r_data*r_data).real
    
    if isinstance(mask,np.ndarray):
        log.info(mask.dtype)
        mask = ~mask
        mask = mask.astype(float)
        mask = np.roll(np.roll(mask,shift[0],axis = 0),shift[1],axis=1) # correct off center origin for point inversion
        mask = np.pad(mask,((n_x_points,0),(n_y_points,0)))
        r_mask = np.fft.fft2(mask)
        mask_convolution =  np.fft.ifft2(r_mask*r_mask).real
        non_zero_mask = mask_convolution!=0
        convolution[non_zero_mask] /= mask_convolution[non_zero_mask]
    log.info('convolution shape ={}'.format(convolution.shape))

    x_len=grid[:,0,0].max()-grid[:,0,0].min()
    y_len=grid[0,:,1].max()-grid[0,:,1].min()
    x_step=np.abs(grid[1,0,0] - grid[0,0,0])
    y_step=np.abs(grid[0,1,1] - grid[0,0,1])
    log.info("len/step x: {} , y:{}".format(x_len/x_step,y_len/y_step))

    n_conv = np.floor((2*x_len/x_step+2 +0.5,2*y_len/y_step+2 +0.5))
    conv_x = -x_len-x_step + np.arange(n_conv[0])*x_step
    conv_y = -y_len-y_step + np.arange(n_conv[1])*y_step
    conv_grid = GridFactory.construct_grid('uniform',[conv_x,conv_y])
    log.info('conv_grid shape ={}'.format(conv_grid.shape))

    #log.info('argmax convolution ={}'.format(np.argmax(convolution)))
    max_index = np.unravel_index(np.argmax(convolution),convolution.shape)
    #log.info('max index ={}'.format(max_index))
    max_grid_point = conv_grid.array[max_index]
    log.info('center ={}'.format(max_grid_point))

    max_area_sigma_mask = (convolution > convolution.max()*np.exp(-.5))    
    sigma_grid_points = conv_grid[max_area_sigma_mask,:]-max_grid_point    
    log.info('sigma gridpoints shape = {}'.format(sigma_grid_points.shape))
    if sigma_grid_points.shape[0] != 0:
        one_sigma_deviation = np.array(((sigma_grid_points[...,0].min(),sigma_grid_points[...,0].max()),(sigma_grid_points[...,1].min(),sigma_grid_points[...,1].max())))
    else:
        one_sigma_deviation = np.array([[np.inf, np.inf],[np.inf,np.inf]])


    out_dict = {}
    out_dict['convolution'] = convolution
    out_dict['convolution_grid'] = conv_grid
    out_dict['max'] = max_grid_point
    out_dict['center'] = max_grid_point/2
    out_dict['one_sigma_mask'] = max_area_sigma_mask    
    out_dict['one_sigma_deviations'] = one_sigma_deviation/2
    return out_dict
    

def find_center_of_mass(data,points):
    dim = points.shape[-1]
    return np.mean(data.flatten()*points.reshape(-1,dim) )/np.sum(data)



import numpy as np
import logging
import vtk
from vtk.util import numpy_support as vtn
from xframe.library.gridLibrary import GridFactory
from xframe.library.gridLibrary import NestedArray
from xframe.database.interfaces import VTKInterface
log=logging.getLogger('root')

class vtkSaver(VTKInterface):
    lib = vtk    
    dset_names_default = 'data'
    default_grid = False
    file_type_dict={'cartesian':'vtr','spherical':'vts','polar':'vts'}
    @staticmethod
    def correct_file_path_ending(path,grid_type='cartesian'):
        old_path = path
        file_type = vtkSaver.file_type_dict.get(grid_type,'vtr')
        ending = path.split('.')[-1]
        if len(ending) == 3:
            path = path[:-3] + file_type
        else:
            path = path + '.'+ file_type
        #log.info('grid_type = {}\n path = {} \n corrected path = {} \n file_type = {}'.format(grid_type,old_path,path,file_type))
        return path
    @staticmethod
    def save(file_path,datasets, grid=default_grid,dset_names = dset_names_default, grid_type='cartesian',**kwargs):
        '''
        Saves vtk 3d models which can be opened e.g with Paraview.
        Currenty the types 'cartesian' and 'spherical' are supported.
        They correspond to uniform cartesian and sperical grids.
        Throws assertion error if unknown grid type is provided.
        '''
        try:
            assert len(datasets)>0,'No datasets provided, i.e.: len(datasets)=0'
            writer = False
            for t,get_writer in vtkSaver.writer_dict.items():
                if t==grid_type:
                    writer = get_writer(datasets,grid = grid,dset_names=dset_names)
                    file_path = vtkSaver.correct_file_path_ending(file_path,grid_type=grid_type)
                    break
            assert ~isinstance(writer,bool),'grid_type {} not known. Known grid_types are {}'.format(grid_type,vtkSaver.writer_dict.keys())

            
            writer.SetFileName(file_path)
            writer.Update()
        except AssertionError as e:
            log.error(e)
            
    @staticmethod
    def load(path,**kwargs):
        e=NotImplementedError('loading of VTK filetypes is not supported yet.')
        raise e
            


    @staticmethod
    def generate_spherical_writer( datasets, grid=default_grid, dset_names=dset_names_default):
        for n,data in enumerate(datasets):
            new_data = np.zeros((data.shape[0],)+(data.shape[1]+2,)+(data.shape[-1]+1,))
            new_data[:,1:-1,:-1] = data
            new_data[:,1:-1,-1] = data[:,:,0]
            new_data[:,0,:-1] = np.mean(data[:,0,:],axis = 1)[:,None]
            new_data[:,-1,:-1] = np.mean(data[:,-1,:],axis = 1)[:,None]
            datasets[n]=new_data
        if not (isinstance(grid,np.ndarray) or isinstance(grid,NestedArray)):
            n_r,n_theta,n_phi = datasets[0].shape
            rs = np.arange(n_r)*1/n_r
            phis = np.arange(n_phi-1)*np.pi*2/(n_phi-1)
            thetas = (np.arange(n_theta)*np.pi/(n_theta-1))[1:-1]
            grid = GridFactory.construct_grid('uniform',[rs,thetas,phis])
        phis = np.zeros(len(grid[0,0,:,2])+1,grid[0,0,:,2].dtype)
        phis[:-1]=grid[0,0,:,2]
        phis[-1] = 0 # periodic point
        thetas = np.zeros(len(grid[0,:,0,1])+2,grid[0,:,0,1].dtype)
        thetas[1:-1]=grid[0,:,0,1]
        thetas[0]=0
        thetas[-1]=np.pi
        grid= np.stack(np.meshgrid(phis,thetas,grid[:,0,0,0],indexing='ij'),3) 
        grid[...,1]-=np.pi/2
        xy_projection= grid[...,2]*np.cos(grid[...,1])
        x=xy_projection*np.cos(grid[...,0])
        y=xy_projection*np.sin(grid[...,0])
        z=grid[...,2]*np.sin(grid[...,1])
    
        points = np.empty((x.size, 3), x.dtype)
        points[:, 0] = x.ravel('F')
        points[:, 1] = y.ravel('F')
        points[:, 2] = z.ravel('F')


        points = np.require(points, requirements=['C'])
        vtkpts = vtk.vtkPoints()
        vtk_arr = vtn.numpy_to_vtk(points, deep=True)
        vtkpts.SetData(vtk_arr)

        vtk_grid = vtk.vtkStructuredGrid()
        vtk_grid.SetDimensions(*grid.shape[:3])
        vtk_grid.SetPoints(vtkpts)
    
        for n,density in enumerate(datasets):
            density = np.swapaxes(density.copy(),0,2)
            density = np.require(density, requirements=['C']).real
            vtk_density = vtn.numpy_to_vtk(num_array=np.swapaxes(density,0,2).copy().ravel(), deep=True, array_type=vtk.VTK_FLOAT)    
            vtk_grid.GetPointData().AddArray(vtk_density)
            # print grid.GetPointData().GetNumberOfArrays()
            if isinstance(dset_names,str):
                vtk_density.SetName("{}_{}".format(dset_names,n))
            elif isinstance(dset_names,(list,tuple)):
                vtk_density.SetName("{}_{}".format(dset_names[n],n))                        
        
        writer = vtk.vtkXMLStructuredGridWriter()
        writer.SetInputData(vtk_grid)
        return writer


    @staticmethod
    def generate_polar_vtkStructuredGrid(datasets, grid=default_grid, dset_names=dset_names_default):
        for n,data in enumerate(datasets):
            new_data = np.zeros((data.shape[0],)+(1,)+(data.shape[1]+1,))
            new_data[:,0,:-1] = data
            new_data[:,0,-1] = data[:,0]
            datasets[n]=new_data
        if not (isinstance(grid,np.ndarray) or isinstance(grid,NestedArray)):
            n_r,n_phi = datasets[0].shape
            n_theta = 1
            rs = np.arange(n_r)*1/n_r
            phis = np.arange(n_phi-1)*np.pi*2/(n_phi-1)
            thetas = np.array([np.pi/2])
            grid = GridFactory.construct_grid('uniform',[rs,thetas,phis])
        phis = np.zeros(len(grid[0,:,1])+1,grid[0,:,1].dtype)
        phis[:-1]=grid[0,:,1]
        phis[-1] = 0 # periodic point
        thetas = np.array([np.pi/2])
        #thetas[-1]=np.pi
        grid= np.stack(np.meshgrid(phis,thetas,grid[:,0,0],indexing='ij'),3) 
        grid[...,1]-=np.pi/2
        xy_projection= grid[...,2]*np.cos(grid[...,1])
        x=xy_projection*np.cos(grid[...,0])
        y=xy_projection*np.sin(grid[...,0])
        z=grid[...,2]*np.sin(grid[...,1])
    
        points = np.empty((x.size, 3), x.dtype)
        points[:, 0] = x.ravel('F')
        points[:, 1] = y.ravel('F')
        points[:, 2] = z.ravel('F')


        points = np.require(points, requirements=['C'])
        vtkpts = vtk.vtkPoints()
        vtk_arr = vtn.numpy_to_vtk(points, deep=True)
        vtkpts.SetData(vtk_arr)

        vtk_grid = vtk.vtkStructuredGrid()
        vtk_grid.SetDimensions(*grid.shape[:3])
        vtk_grid.SetPoints(vtkpts)
    
        for n,density in enumerate(datasets):
            density = np.swapaxes(density.copy(),0,2)
            density = np.require(density, requirements=['C']).real
            vtk_density = vtn.numpy_to_vtk(num_array=np.swapaxes(density,0,2).copy().ravel(), deep=True, array_type=vtk.VTK_FLOAT)    
            vtk_grid.GetPointData().AddArray(vtk_density)
            # print grid.GetPointData().GetNumberOfArrays()
            if isinstance(dset_names,str):
                vtk_density.SetName("{}_{}".format(dset_names,n))
            elif isinstance(dset_names,(list,tuple)):
                vtk_density.SetName("{}_{}".format(dset_names[n],n))
            else:
                vtk_density.SetName("data_{}".format(n))
        return vtk_grid
        
    @staticmethod
    def generate_polar_writer( datasets, grid=default_grid, dset_names=dset_names_default):
        vtk_grid = vtkSaver.generate_polar_vtkStructuredGrid(datasets, grid=grid, dset_names=dset_names)
        writer = vtk.vtkXMLStructuredGridWriter()
        writer.SetInputData(vtk_grid)
        return writer
    
    
    
    @staticmethod
    def generate_cartesian_writer(datasets, grid=default_grid, dset_names=dset_names_default):
        if not (isinstance(grid,np.ndarray) or isinstance(grid,NestedArray)):
            shape = datasets[0].shape
            axes = [np.arange(ax_len) for ax_len in shape]
            grid = GridFactory.construct_grid('uniform',axes).array
        g_shape  = grid[:].shape[:-1]
        n_dim = len(g_shape)
        if n_dim == 2:
            g_shape = g_shape + (1,)
        vtk_grid = vtk.vtkRectilinearGrid()
        vtk_grid.SetDimensions(g_shape)

        if n_dim == 2:
            vtk_grid.SetXCoordinates(vtn.numpy_to_vtk(grid[:,0,0],deep = True))
            vtk_grid.SetYCoordinates(vtn.numpy_to_vtk(grid[0,:,1],deep = True))
        if n_dim == 3:
            vtk_grid.SetXCoordinates(vtn.numpy_to_vtk(grid[:,0,0,0],deep = True))
            vtk_grid.SetYCoordinates(vtn.numpy_to_vtk(grid[0,:,0,1],deep = True))
            vtk_grid.SetZCoordinates(vtn.numpy_to_vtk(grid[0,0,:,2],deep = True))
            
        for n,data in enumerate(datasets):
            vtk_array = vtn.numpy_to_vtk(data.ravel('F'),deep = True ,array_type = vtk.VTK_FLOAT)
            vtk_grid.GetPointData().AddArray(vtk_array)
            if isinstance(dset_names,str):
                vtk_array.SetName("{}_{}".format(dset_names,n))
            elif isinstance(dset_names,(list,tuple)):
                vtk_array.SetName("{}_{}".format(dset_names[n],n))                        
        writer = vtk.vtkXMLRectilinearGridWriter()
        writer.SetInputData(vtk_grid)
        return writer        


    # writer_dict definition needs to be placed after the definitions of generate_(cartesian/spherical)_writer 
    writer_dict={'cartesian':generate_cartesian_writer.__func__,'spherical':generate_spherical_writer.__func__,'polar':generate_polar_writer.__func__}    

    



import os
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
import numpy as np
from xframe.detectors.agipd import AGIPD
from xframe.library import physicsLibrary as pLib
from xframe.library.pythonLibrary import grow_mask
from xframe.presenters.matplolibPresenter import agipd_heat
from xframe.presenters.matplolibPresenter import heat2D
from xframe.startup_routines import dependency_injection,load_recipes
from xframe.library.mathLibrary import cartesian_to_spherical

init = False
if init:
    dependency_injection()
    from xframe.control.Control import Controller
    analysis_sketch,exp_sketch = load_recipes(ana_name='fxs3046_online.generate_plots_2', ana_settings='mean_test',exp_name='fxs3046_online',exp_settings = 'default_proc')
    controller = Controller(analysis_sketch,experiment_sketch=exp_sketch)
    controller.chooseJob()
    from scipy import ndimage
    from xframe import settings
    from xframe import Multiprocessing
    from xframe import database

db = database.analysis
ex_db = database.experiment
comm = Multiprocessing.comm_module

opt = settings.analysis
ex_opt = settings.experiment

def setup_locals():    
    agipd = AGIPD(ex_db,load_geometry_file=True)
    agipd.origin= ex_opt.detector_origin
    energy = ex_opt.x_ray_energy
    wavelength = pLib.energy_to_wavelength(energy)
    geometry = comm.get_geometry()
    return locals()


#locals().update(setup_locals())


def feature_selection(mask,asic_border_mask):
    mask = (mask.astype(int) - asic_border_mask.astype(int)).astype(bool)
    main_mask = np.full(mask.shape,0).astype(int)
    #main_mask[:]=mask
    small_features = np.full(mask.shape,0)
    for m in range(16):
        m_labels,m_nf = ndimage.label(mask[m])
        for l in range(1,m_nf):
            feature_mask = (m_labels == l)
            n_masked_pixels = np.sum(m_labels==l)        
            if n_masked_pixels >3 and n_masked_pixels <1e4  :
                tmp_mask = main_mask[m]
                tmp_mask += 2*l*feature_mask
                #main_mask[m]=tmp_mask
            elif  n_masked_pixels <=3:
                tmp_mask = small_features[m]
                tmp_mask += l*feature_mask
                #small_features[m]=tmp_mask
    return main_mask, small_features

def select_bad_pixels(fig_dict,limits={'std':[5,None],'max':[60,300]}):
    globals().update(setup_locals())
    std_dict = fig_dict['std_2d_generic']
    std_data = np.mean([std_dict[key] for key in std_dict],axis =0)
    max_data = fig_dict['maximum_2d']


    tmp_mask = ((std_data*max_data)!=0)
    bare_mask = np.array(agipd.sensitive_pixel_mask)
    bare_mask[bare_mask] = tmp_mask.flatten()
    bare_mask = ~bare_mask

    #exclude points with max lower than on all modules:
    if isinstance(limits['max'][0],(int,float)):
        max_data[max_data<limits['max'][0]]=0
        
    #exclude points with max higher than on some modules:
    for m in [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15]:
        tmp_data = max_data[m]
        if isinstance(limits['max'][1],(int,float)):            
            tmp_data[tmp_data>limits['max'][1]]=0
        max_data[m] = tmp_data

    #mask things that are zero
    tmp_mask = (max_data!=0).astype(float)
    mask_max = np.array(agipd.sensitive_pixel_mask)
    mask_max[mask_max] = tmp_mask.flatten()
    mask_max = ~mask_max
        
    #exclude points with std higher/lower than on some modules:
    for m in [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15]:
        tmp_data = std_data[m]
        if isinstance(limits['std'][0],(int,float)):
            tmp_data[tmp_data<limits['std'][0]]=0
        if isinstance(limits['std'][1],(int,float)):
            tmp_data[tmp_data>limits['std'][1]]=0
        std_data[m] = tmp_data

    #mask things that are zero
    tmp_mask = (max_data!=0).astype(float)
    mask_std = np.array(agipd.sensitive_pixel_mask)
    mask_std[mask_std] = tmp_mask.flatten()
    mask_std = ~mask_std

    return mask_max,mask_std


def get_border_mask(data_shape):
    border_mask = np.full(data_shape,False)
    border_mask[:,:2]=True
    border_mask[:,62:67]=True
    border_mask[:,127:132]=True
    border_mask[:,192:197]=True
    #border_mask[:,257:264]=True
    border_mask[:,257:262]=True 
    border_mask[:,322:327]=True
    border_mask[:,387:392]=True
    border_mask[:,452:457]=True
    border_mask[:,-2:]=True
    border_mask[:,:,-1:]=True
    border_mask[:,:,:1]=True
    return border_mask


def get_manual_mask(data_shape):
     # bad_asics_in module 15
    m_15_mask = np.full(data_shape,False)
    m_15_mask[15,:197]=True

    # bad_asic_in module 5
    m_5_mask = np.full(data_shape,False)
    m_5_mask[5,262:322,:64]=True

    # bad_asic_in module 10
    m_10_mask = np.full(data_shape,False)
    m_10_mask[10,262:322,64:]=True

    # bad_asic_in module 13
    m_13_mask = np.full(data_shape,False)
    m_13_mask[13,262:322,:64]=True
    
    # bad_asics_in module 1
    m_0_mask = np.full(data_shape,False)
    m_0_mask[0,:67,30:50]=True
    m_0_mask[0,457:,64:]=True

    # bad_asics_in module 3
    m_3_mask = np.full(data_shape,False)
    m_3_mask[3,:67,62:82]=True

    # bad_asics_in module 11
    m_11_mask = np.full(data_shape,False)
    m_11_mask[11,327:392,16:32]=True

    combined = m_15_mask | m_0_mask | m_3_mask | m_11_mask | m_5_mask | m_10_mask | m_13_mask

    return {'combined':combined,'module_15':m_15_mask,'module_0':m_0_mask,'module_3':m_3_mask,'module_11':m_11_mask,'module_5':m_5_mask,'module_10':m_10_mask,'module_13':m_13_mask}

def reformat_masks(masks,sensitive_pixel_mask,data_shape):
    ags=sensitive_pixel_mask
    ds = data_shape
    for key in masks:
        if isinstance(masks[key],dict):
            masks[key]=reformat_masks(masks[key],sensitive_pixel_mask,data_shape)
        else:
            masks[key] = ~masks[key].astype(bool)[ags].reshape(ds)
    return masks

def combine_masks(high_masks,medium_masks):
    globals().update(setup_locals())
    data_shape = high_masks[0].shape

    high_gain_mask = high_masks[0] * high_masks[1]
    medium_gain_mask = medium_masks[0] * medium_masks[1]

    max_mask = high_masks[0] | medium_masks[0]
    std_mask = high_masks[1] | medium_masks[1]
    
    border_mask = get_border_mask(data_shape)

    mask_c = max_mask*std_mask 
    main_mask,small_features = feature_selection(mask_c.copy(),border_mask)
    #main_mask_std,small_std = feature_selection(mask_std,border_mask)

    #small_features = small_max & small_std
    #main_mask = main_mask_max & main_mask_std

    manual_masks = get_manual_mask(data_shape)
    combined_manual = manual_masks.pop('combined')
    enlarged_main_mask = grow_mask(main_mask,2)

    combined_mask= (main_mask.astype(bool) | small_features.astype(bool) | border_mask.astype(bool) | combined_manual.astype(bool) | enlarged_main_mask.astype(bool))

    masks = {'combined':combined_mask,'main_fatures':main_mask,'small_features':small_features,'enlarged_main_features':enlarged_main_mask,'asic_border_mask':border_mask,'manual_masks':manual_masks,'high_gain_mask':high_gain_mask,'medium_gain_mask':medium_gain_mask}

    ags=agipd.sensitive_pixel_mask
    ds = (16,512,128)
    masks = reformat_masks(masks,ags,ds)
    return masks
    

def generation_routine(medium_dict,high_dict,limit={'std':[5,None],'max':[60,300]}):
    globals().update(setup_locals())
    
    
    detected_values = (mask_max*mask_std).copy()
    #mask[:] = 1


            #mask = ~mask#mask = ~mask

    border_mask = np.full(mask_max.shape,False)
    border_mask[:,:2]=True
    border_mask[:,62:67]=True
    border_mask[:,127:132]=True
    border_mask[:,192:197]=True
    #border_mask[:,257:264]=True
    border_mask[:,257:262]=True 
    border_mask[:,322:327]=True
    border_mask[:,387:392]=True
    border_mask[:,452:457]=True
    border_mask[:,-2:]=True
    border_mask[:,:,-1:]=True
    border_mask[:,:,:1]=True


    # bad_asics_in module 15
    m_15_mask = np.full(mask_max.shape,False)
    m_15_mask[15,:197]=True

    # bad_asics_in module 1
    m_0_mask = np.full(mask_max.shape,False)
    m_0_mask[0,:67,30:50]=True
    m_0_mask[0,457:,64:]=True

    # bad_asics_in module 3
    m_3_mask = np.full(mask_max.shape,False)
    m_3_mask[3,:67,62:82]=True

    # bad_asics_in module 11
    m_11_mask = np.full(mask_max.shape,False)
    m_11_mask[11,327:392,16:32]=True
    

    #bar_plus_mask = np.full(mask.shape,False)
    #bar_plus_mask[4:8,255:257]=True


    mask_std = (mask_std.astype(int) - border_mask.astype(int)).astype(bool)



    main_mask_max,small_max = feature_selection(mask_max)
    main_mask_std,small_std = feature_selection(mask_std)

    small_features = small_max & small_std
    main_mask = main_mask_max & main_mask_std
    
    main_mask[border_mask] = 0
    #border_mask -= agipd_heat.sensitive_pixel_mask 
    #grads= ndimage.sobel(main_mask[0].astype(int))
    main_mask_int = main_mask.copy()
    main_mask = main_mask
    grads1= np.sum(np.abs(np.gradient(main_mask.astype(int),edge_order = 1)[1:]),axis = 0)
    grads2= np.sum(np.abs(np.gradient((grads1!=0).astype(int),edge_order = 1)[1:]),axis = 0)
    new_mask = (grads1!=0) | (grads2!=0)

    #xy_grads = np.sum(np.array(grads[1:]),axis = 0)

    #show_mask = m_labels
    ms=15
    #heat2D.show(main_mask[0]+2*new_mask,agipd.pixel_grid[ms],scale='lin')
    #agipd_heat.show(np.abs(2*final_mask-new_mask)[agipd.sensitive_pixel_mask],agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'},scale='lin',cmap_name='viridis_r')
    
    ags=agipd.sensitive_pixel_mask
    ds = (16,512,128)
    main_mask = main_mask.astype(bool)[ags].reshape(ds)
    main_mask_int = main_mask_int[ags].reshape(ds)
    small_features = small_features.astype(bool)[ags].reshape(ds)
    new_mask = new_mask.astype(bool)[ags].reshape(ds)
    border_mask = border_mask.astype(bool)[ags].reshape(ds)
    #bar_mask = bar_mask.astype(bool)[ags].reshape(ds)
    #bar_plus_mask =  bar_plus_mask.astype(bool)[ags].reshape(ds)
    m_15_mask =  m_15_mask.astype(bool)[ags].reshape(ds)
    m_0_mask =  m_0_mask.astype(bool)[ags].reshape(ds)
    m_3_mask =  m_3_mask.astype(bool)[ags].reshape(ds)
    m_11_mask =  m_11_mask.astype(bool)[ags].reshape(ds)
    detected_values = detected_values.astype(bool)[ags].reshape(ds)
    

    combined_mask = ~(main_mask | small_features | new_mask |border_mask | bar_mask | bar_plus_mask)

    mask_dict = {'mask':combined_mask,'main_features':main_mask,'small_features':small_features,'enlagement_main':new_mask,'asic_borders':border_mask,'bars':bar_mask,'bar_asic_gap':bar_plus_mask,'main_features_int':main_mask_int,'module_15_bad_asics':m_15_mask,'module_0_bad_asics':m_0_mask,'module_3_bad_asic_area':m_3_mask,'module_11_bad_asic_area':m_11_mask,'detected_values':detected_values}
    globals().update(locals())
    return mask_dict

#stupid_limits = {'std':[66,60],'max':[310,300]} 

#fig_dict_low = db.load('figure_data',path_modifiers={'name':'bg_data_proc','run':3,'time':'2022-08-31_20:47:52'})
#low_limits = {'std':[100,1e4],'max':[4000,1e5]}
#mask_dict_low = generation_routine(fig_dict_low,limits = low_limits)#low_limits)

fig_dict_high = db.load('figure_data',path_modifiers={'name':'bg_data_proc','run':1,'time':'2022-08-31_20:45:39'})
fig_dict_medium = db.load('figure_data',path_modifiers={'name':'bg_data_proc','run':2,'time':'2022-08-31_20:46:53'})
medium_limits = {'std':[200,2000],'max':[2000,2e4]}
high_limits = {'std':[6,60],'max':[25,300]}

medium_masks = select_bad_pixels(fig_dict_medium,limits = medium_limits)
high_masks = select_bad_pixels(fig_dict_high,limits = high_limits)

mask_dict = combine_masks(high_masks,medium_masks)

#mask_dict = database.experiment.load('custom_mask',path_modifiers={'name':'agipd_mask'})
fig = agipd_heat.get_fig(mask_dict['combined'],agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'Agipd Mask p3046','y_label':'y_direction [mm]','x_label':'x_direction [mm]'},scale='lin')
db.save('/gpfs/exfel/theory_group/user/berberic/p3046/masks/combined.matplotlib',fig,dpi=800)
db.save('/gpfs/exfel/theory_group/user/berberic/p3046/masks/new_mask.h5',mask_dict)
#fig = agipd_heat.get_fig(~mask_dict['high_gain_mask'],agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'},scale='lin')
#db.save('/gpfs/exfel/theory_group/user/berberic/p3046/masks/high_gain.matplotlib',fig,dpi=800)
#fig = agipd_heat.get_fig(~mask_dict['medium_gain_mask'],agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'},scale='lin')
#db.save('/gpfs/exfel/theory_group/user/berberic/p3046/masks/medium_gain.matplotlib',fig,dpi=800)
#fig = agipd_heat.get_fig(~mask_dict['enlarged_main_features'],agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'},scale='lin')
#db.save('/gpfs/exfel/theory_group/user/berberic/p3046/masks/test.matplotlib',fig,dpi=800)





#mask_dict = db_ana.load_default({'name':'/gpfs/exfel/exp/SPB/202102/p002510/scratch/berberich/mask.h5'},use_name_as_path = True)

#streak_mask = np.full((16,512,128),True)
#streak_mask[4,:,:94]=False
#streak_mask[-1,:,:94]=False

#mask_low = ~ (mask_dict_low['asic_borders'] |  mask_dict_low['enlagement_main'] | mask_dict_low['main_features'] | mask_dict_low['small_features'])
#mask_low = ~ (mask_dict_low['main_features'])
#mask_medium = ~ (mask_dict_medium['asic_borders'] | mask_dict_medium['enlagement_main'] | mask_dict_medium['main_features'] | mask_dict_medium['small_features'])
#mask_medium = ~ (mask_dict_medium['detected_values'])
#mask_high = ~ (mask_dict_high['asic_borders'] | mask_dict_high['enlagement_main'] | mask_dict_high['main_features'] | mask_dict_high['small_features'] | mask_dict_high['module_15_bad_asics'] | mask_dict_high['module_0_bad_asics'] | mask_dict_high['module_3_bad_asic_area']| mask_dict_high['module_11_bad_asic_area'])
#mask_high = ~ (mask_dict_high['detected_values'] )

#mask_or =  mask_high | mask_medium
#mask_and =  mask_high & mask_medium
#
#slope = 128/(512+100)
#for x in range(512):
#    y=128-int(slope *x)
#    streak_mask[4,x,:y]=False
#    streak_mask[-1,x,:y]=False
#    
#mask =  mask & streak_mask
#
#mask_dict['crude_steak_mask']=streak_mask
#mask_dict['mask'] = mask




#db_ana.save_default(mask_dict,{'name':'/gpfs/exfel/exp/SPB/202102/p002510/scratch/berberich/mask.h5'},use_name_as_path = True)


#plot_mask = base
#plot_mask[8] += normal_mask[8]==new_mask[8]
#plot_mask= (new_mask)
#fig = agipd_heat.get_fig(mask_high,agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'},scale='lin')
#db.save('/gpfs/exfel/theory_group/user/berberic/p3046/masks/high_mask.matplotlib',fig,dpi=800)
#fig = agipd_heat.get_fig(mask_medium,agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'},scale='lin')
#db.save('/gpfs/exfel/theory_group/user/berberic/p3046/masks/medium_mask.matplotlib',fig,dpi=800)
##fig = agipd_heat.get_fig(mask_low,agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'},scale='lin')
##db.save('/gpfs/exfel/theory_group/user/berberic/p3046/masks/low_mask.matplotlib',fig,dpi=800)
#fig = agipd_heat.get_fig(mask_or,agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'},scale='lin')
#db.save('/gpfs/exfel/theory_group/user/berberic/p3046/masks/or_mask.matplotlib',fig,dpi=800)
#fig = agipd_heat.get_fig(mask_and,agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'},scale='lin')
#db.save('/gpfs/exfel/theory_group/user/berberic/p3046/masks/and_mask.matplotlib',fig,dpi=800)

#fig = agipd_heat.get_fig(bar_mask[agipd.sensitive_pixel_mask],agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'},scale='lin',cmap_name='viridis_r')


#fig.savefig('/gpfs/exfel/exp/SPB/202102/p002510/scratch/berberich/figures/tmp_mask.png',dpi=800)

#fig.show()

# fig = agipd_heat.get_fig(std_data,agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'std'})
# fig.savefig('/gpfs/exfel/theory_group/user/berberic/gain_test/figures/Standard_Deviation_2D/run_164/std.png',dpi = 800)
# fig = agipd_heat.get_fig(max_data,agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'max'})
# fig.savefig('/gpfs/exfel/theory_group/user/berberic/gain_test/figures/Maximum 2D/run_164/max.png',dpi = 800)
# fig = agipd_heat.get_fig(std_data*max_data,agipd.pixel_grid,agipd.sensitive_pixel_mask,layout={'title':'combined'})
# fig.savefig('/gpfs/exfel/theory_group/user/berberic/gain_test/figures/Standard_Deviation_2D/run_164/combined.png',dpi = 800)


#heat2D.show(d['mean'],grid = regridder.regrid_data['cart']['lab_small']['grid'],scale='log')

#custom_mask = np.full(agipd.data_shape,True).astype(bool)
#db.save('custom_mask',{'mask':custom_mask},path_modifiers = {'name':'custom_mask'})

#background =  np.full(agipd.data_shape,0)
#background[4] = -1000.4
#db.save('background',{'background':background},path_modifiers = {'name':'background'})




import logging
import numpy as np
#from numba import njit
import traceback
import abc
from xframe.library.pythonLibrary import grow_mask
from xframe.library.pythonLibrary import DictNamespace
from xframe.library.mathLibrary import masked_mean
from xframe.library.pythonLibrary import convert_to_slice_if_possible
from xframe.library.mathLibrary import combine_means
from xframe.library.mathLibrary import masked_std
from xframe import Multiprocessing

import warnings


#############################################
# Note in the present analysis and experimen workers filters are applied inside child processes
# so don't use multiprocessing in their apply routines exept if you know exactly what you are doing.
#
# That is alsy why you need to modify the 'data' and 'mask' arrays of a data_chunk in place,
# since they are views in a shared array.
# That means you need to do e.g. data_chunk['data'][:]=0 instead of data_chunk['data'] = np.zeros_like(data_chunk['data'])

log = logging.getLogger('root')

class FilterTools():
    ''' Class that contins static methods that might be important for a single filter as well as a sequence of fliters.'''
    @staticmethod
    def init_masks(filtered_mask=None,modified_mask=None):
        return {'total_filtered':filtered_mask,'total_modified': modified_mask, 'filtered':[filtered_mask] , 'modified':[modified_mask]}    
    
    @staticmethod
    def combine_masks(masks, filtered_mask, modified_mask):
        if len(masks)!=4:
            masks = FilterTools.init_masks(filtered_mask,modified_mask)
        else:                       
            masks['filtered'].append(filtered_mask)
            masks['modified'].append(modified_mask)
            #masks['total_modified'][~masks['total_filtered']] = modified_mask.flatten()
            masks['total_filtered'] |= filtered_mask.flatten()
        return masks
    
    @staticmethod
    def remove_filtered_elements(data_chunk,filtered_mask):
        '''
        Assumes elements of data_chuk are filtered along their first dimension. That means it assumes all masks are 1 dimensional.
        filtered_mask is 1 for data values to exclude and 0 for data values to keep. 
        '''
        # If masks dont seem to be present do nothing
        if np.atleast_1d(filtered_mask).any():
            len_data = len(filtered_mask)
            for key,item in data_chunk.items():
                if isinstance(item, (np.ndarray,tuple,list)):
                    if len(item) == len_data:
                        log.info('replacing: {}'.format(key))
                        data_chunk[key] = item[~filtered_mask]
            log.info('alive')
            n_good_frames = len_data - np.sum(filtered_mask)
            log.info('{} % or {} of frames remain after filtering'.format(n_good_frames/len_data*100,n_good_frames))
        return data_chunk
    
class FilterSequence(FilterTools):
    '''
    Allows to combine list of filters into a single object that can be applied to a data_chunk.
    '''
    def __init__(self,filter_list):        
        self.filter_list = filter_list
    def apply(self,data_chunk,masks={}):
        filter_result = [data_chunk,masks]
        for _filter in self.filter_list:
            #log.info('apply {}'.format(_filter))
            filter_result = _filter.apply(*filter_result)
        return filter_result
    
    def append_filter(self,_filter):
        self.filter_list.append(_filter)        
    def get_data(self):
        f_data ={f.name:f.data for f in self.filter_list}
        return f_data
    def reset_data(self):
        for f in self.filter_list:
            f.reset_data()
class Filter(abc.ABC,FilterTools):
    '''
    Abstract filter class from which all custimiced filters need to inherit.
    Logic is as follows each Filter needs to have a function method _apply(data_chunk:dict,masks:dict), that does the following:
    It does something to data_chunk and records in two masks which elements where modified by the filter and which where filtered (i.e. excluded).
    It then needs to return: new_data_chunk, filtered_mask, modified_mask.

    The filter itself handles the cimbination of the filtered and modified masks.
    '''
    # filtered_mask is true when the data is supposed to be excluded (i.e. was filtered out)
    # modified_mask is true when the data was modified  
    def __init__(self, opt : dict):
        self.opt = opt
        self.name = opt['name']
        self.data = {}
        self.roi_mask = opt['roi_mask']
        #log.info('roi_mask shape = {} any = {}'.format(self.roi_mask.shape,self.roi_mask.any()))
        #log.info('roi_modules = {}'.format(self.roi_modules))
        
    @abc.abstractmethod
    def _apply(self, data_chunk : dict, masks : dict):
        pass

    def reset_data(self):
        self.data={}
                
    #returns new calib_chunk and masks
    def apply(self, calib_chunk : dict , masks : dict = {}):
        filtered_chunk, filtered_mask,modified_mask = self._apply(calib_chunk, masks)
        new_masks = self.combine_masks(masks,filtered_mask,modified_mask)
        return filtered_chunk,new_masks

    
class BraggFilter(Filter):
    def __init__(self, *args):
        super().__init__(*args)
        self.mode = self.opt.get('mode','discard') # 'mask'
        self.max_sigma = self.opt.get('max_sigma',3)
        self.radial_bin_size_in_pixels = self.opt.get('radial_bin_size_in_pixels',3)
        self.pixel_grid = self.opt['data_grid']
        self.q_limits = self.opt.get('q_limits',[None,None])        
        self.fill_value = self.opt.get('fill_value',0)
        self.use_log_scale = self.opt.get('log_scale',False)
        self.radial_overlap_in_pixels = self.opt.get('radial_overlap_in_pixels',0)
        self.max_std_growth = self.opt.get('max_std_growth',0)
        self.min_frames_per_mean = self.opt.get('min_part_len',70)
        
        if self.radial_overlap_in_pixels >= self.radial_bin_size_in_pixels:
            self.radial_overlap_in_pixels = 1
            log.error('Radial overlap bigger or equal to radial bin size setting overlap to 1 pixel')
        
        # if true will calculate the number of 
        self.save_bragg_pixel_count = self.opt.get('save_bragg_pixel_count',False)
        self.pixel_width = abs(self.pixel_grid[0,-1,0,0] - self.pixel_grid[0,-2,0,0])
        #log.info('pixel width = {}'.format(self.pixel_width))
        self.radial_masks, self.qs = self.generate_growing_radial_masks()
    

    def generate_growing_radial_masks(self):
        # important that the masks are ordered from lowest to highest q-bin
        grid = self.pixel_grid
        qs = np.linalg.norm(grid,axis = -1)        
        step = self.radial_bin_size_in_pixels*self.pixel_width
        overlap = self.radial_overlap_in_pixels*self.pixel_width
        q_limits = [qs.min(),qs.max()]
        for i,limit in enumerate(self.q_limits):
            if limit != None:
                q_limits[i] = limit
        #log.info('q_limits = {}'.format(q_limits))
        #q_splits = np.arange(q_limits[0],q_limits[1]+step,step)
        q_splits = np.arange(q_limits[0],q_limits[1]+step,step)
        #radial_masks = tuple((qs>q_splits[i]) & (qs<=q_splits[i+1]+overlap) for i in range(len(q_splits)-1))
        radial_masks = tuple((qs>=q_splits[i]) & (qs<q_splits[i+1]) for i in range(len(q_splits)-1))
        return radial_masks , qs
        
    
    def _apply(self,calib_chunk, masks):
        data = calib_chunk['data']
        mask = calib_chunk['mask']
        if self.mode == 'discard':
            bragg_counts = self.comm.request_mp_evaluation(self.count_bragg_pixels, mp_type = 'shared_array', out_shape = (len(data),), out_dtype = np.dtype(int), argArrays = [data], const_args = [self.radial_masks,self.max_sigma,self.use_log_scale], callWithMultipleArguments = True)
            filtered_mask = bragg_counts.astype(bool)
            #log.info('bragg_counts = {} '.format(bragg_counts))
            modified_mask = np.zeros(filtered_mask.shape,dtype = bool)
        elif self.mode == 'select':
            bragg_counts = self.comm.request_mp_evaluation(self.count_bragg_pixels, mp_type = 'shared_array', out_shape = (len(data),), out_dtype = np.dtype(int), argArrays = [data], const_args = [self.radial_masks,self.max_sigma,self.use_log_scale], callWithMultipleArguments = True)
            filtered_mask = ~bragg_counts.astype(bool)
            #log.info('bragg_counts = {} '.format(bragg_counts))
            modified_mask = np.zeros(filtered_mask.shape,dtype = bool)
        elif self.mode == 'mask':
            # need to detrministically split data based on chunk ids
            data, bragg_mask = self.comm.request_mp_evaluation(self.mask_bragg_pixels, mp_type = 'shared_array_multi', out_shape = data.shape, out_dtypes = [data.dtype,np.dtype(bool)], argArrays = [data, mask], const_args = [], callWithMultipleArguments = True,split_together = True,min_part_len = self.min_frames_per_mean)
            log.info('bad pixels = {} %'.format(np.sum(bragg_mask)/np.sum(mask)*100))
            flat_mask = bragg_mask.reshape(data.shape[0],-1)
            bragg_counts = np.sum(flat_mask, axis = 1)
            calib_chunk['data'] = data
            calib_chunk['mask'] &= ~bragg_mask
            modified_mask = bragg_counts.astype(bool)
            filtered_mask = np.zeros(modified_mask.shape,dtype = bool)
        #log.info('Bragg filter counts = {}'.format(bragg_counts))   
        if self.save_bragg_pixel_count:
            old_counts = self.data.get('n_bragg_pixels',np.array([]))
            new_counts= np.concatenate((old_counts,bragg_counts))
            self.data['n_bragg_pixels'] = new_counts
        return calib_chunk,filtered_mask, modified_mask


    def get_q_sorted_data(self,data):
        masks = self.radial_masks        
        if self.use_log_scale:
            min_data = abs(data.min())+1
            #assumes min of data is 0
            q_sorted_data = tuple(np.log(min_data+data[:,q_mask]) for q_mask in masks)
        else:
            q_sorted_data = tuple(data[:,q_mask] for q_mask in masks)
        return q_sorted_data


    def generate_thresholds(self,data):
        mean = np.mean(data,axis = 0)
        q_means = tuple(mean[q_mask] for q_mask in masks)



    def mask_bragg_pixels(self,data,data_mask,**kwargs):
        part_len = self.min_frames_per_mean
        n_parts = len(data)//part_len
        part_end_ids = [(i+1)*part_len for i in range(n_parts-1)]
        part_end_ids.append(len(data))
        log.info('part end ids = {}'.format(part_end_ids))
        fill_value = self.fill_value
        #log.info('mean shape = {} mask shape = {}'.format(mean.shape,mean_mask.shape))
        bad_pixel_mask = np.zeros(data.shape,dtype =bool)
        very_bad_pixel_mask = np.zeros(data.shape,dtype =bool)
        
        for part_id in range(n_parts):
            part_slice = slice(part_id*part_len,part_end_ids[part_id])
            previouse_std = np.inf
            part_data = data[part_slice]
            part_data_mask = data_mask[part_slice]
            mean, mean_mask = masked_mean(part_data, return_mask = True, axis = 0, where = part_data_mask)
            for q_id,q_mask in enumerate(self.radial_masks):
                q_data = part_data[:,q_mask]            
                stds,stds_mask = masked_std(q_data,fill_value = fill_value, return_mask = True , axis = 1, where = part_data_mask[:,q_mask])
                q_std = masked_mean(stds, fill_value = fill_value, where = stds_mask)
                q_mean = masked_mean(mean[q_mask], fill_value = fill_value, where = mean_mask[q_mask])
                #limiting growth of std for when going to higher q-bins
                if q_std > previouse_std*self.max_std_growth and previouse_std != 0.0:
                    q_std = previouse_std
                previouse_std = q_std
                #log.info('mean {} std {}'.format(q_mean,q_std))
                thresholds = q_mean + q_std*self.max_sigma
                bad_pixel_mask[part_slice,q_mask] = (q_data < -1*thresholds[0]) | (q_data > thresholds[0])
        #    very_bad_pixel_mask[:,q_mask] = (q_data < -1*thresholds[1]) |  (q_data > thresholds[1])
        #for f_id in range(len(data)):
        #    very_bad_pixel_mask[f_id] = grow_mask(very_bad_pixel_mask[f_id],2)
        #bad_pixel_mask |= very_bad_pixel_mask 
        data[bad_pixel_mask] = fill_value
        #log.info('num of bad pixels = {}'.format(np.sum(bad_pixel_mask)))
        #log.info('num of masked pixels = {}'.format(np.sum(~data_mask)))
        #log.info('bad pixels = {} %'.format(np.sum(bad_pixel_mask)/np.sum(data_mask)*100))
        return data,bad_pixel_mask
        
    def mask_bragg_pixels_old(self,data,data_mask,**kwargs):
        masks = self.radial_masks
        max_sigma = self.max_sigma
        fill_value = self.fill_value
        use_log_scale = self.use_log_scale
        if use_log_scale:
            min_data = abs(data.min())+1
            #assumes min of data is 0
            q_sorted_data = tuple(np.log(min_data+data[:,q_mask]) for q_mask in masks)
        else:
            q_sorted_data = tuple(data[:,q_mask] for q_mask in masks)

        bad_pixel_mask = np.zeros(data.shape,dtype =bool)
        very_bad_pixel_mask = np.zeros(data.shape,dtype =bool)
        for q_id,q_data in enumerate(q_sorted_data):
            #log.info('q_data shape = {}'.format(q_data.shape))
            non_zero_mask = q_data != 0
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore')
                # gives divide by 0 error if q frame is compleatelly masked i.e. zero
                stds = np.std(q_data, axis = 1 , where = non_zero_mask)
                stds[np.isnan(stds)] = 0.0
                means = np.mean(q_data, axis = 1 , where = non_zero_mask)
                means[np.isnan(means)] = 0.0
            abs_q_data = np.abs(q_data)
            bad_pixel_mask[:,masks[q_id]] = abs_q_data > (means + stds*max_sigma[0])[:,None]
            very_bad_pixel_mask[:,masks[q_id]] = abs_q_data > (means + stds*max_sigma[1])[:,None]
            
        for f_id in range(len(data)):
            very_bad_pixel_mask[f_id] = grow_mask(very_bad_pixel_mask[f_id],2)
        bad_pixel_mask |= very_bad_pixel_mask 
        data[bad_pixel_mask] = fill_value
        return data, bad_pixel_mask
    
    def mask_bragg_pixels_inter_q(self,data,**kwargs):
        masks = self.radial_masks
        max_sigma = self.max_sigma
        fill_value = self.fill_value      
        max_inter_q_sigma = self.max_inter_q_sigma
        q_hist_length = self.q_hist_length

        q_sorted_data = self.get_q_sorted_data(data)
        bad_pixel_mask = np.zeros(data.shape,dtype =bool)
        very_bad_pixel_mask = np.zeros(data.shape,dtype =bool)
        mean_hist= np.full((self.q_hist_length,len(data)),np.nan)
        std_hist = np.full((self.q_hist_length,len(data)),np.nan)
        for q_id,q_data in enumerate(q_sorted_data):
            #log.info('q_data shape = {}'.format(q_data.shape))
            non_zero_mask = q_data != 0
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore')
                # gives divide by 0 error if q frame is compleatelly masked i.e. zero
                stds = np.std(q_data, axis = 1 , where = non_zero_mask)
                stds[np.isnan(stds)] = 0.0
                means = np.mean(q_data, axis = 1 , where = non_zero_mask)
                means[np.isnan(means)] = 0.0
                
                neither_nan_nor_zero = ~np.isnan(means) & (means !=0)
                thresholds = np.mean(mean_hist, axis = 0 ,where = neither_nan_nor_zero) + np.std(mean_hist, axis = 0 ,where = neither_nan_nor_zero)*max_inter_q_sigma
            nan_thresholds = np.isnan(thresholds)
            thresholds[nan_thresholds] = means[nan_thresholds]
            invalid_means_mask = means > thresholds
            if invalid_means_mask.any():
                #log.info('\n old: {} \n new {}'.format(mean_hist[0][invalid_means_mask],means[invalid_means_mask]))
                #log.info('WARNING: there are invalid means !')
                pass
            means[invalid_means_mask] = mean_hist[0][invalid_means_mask]
            stds[invalid_means_mask] = std_hist[0][invalid_means_mask]


            mean_hist = np.roll(mean_hist,1)
            std_hist = np.roll(std_hist,1)
            mean_hist[0] = means
            std_hist[0] = stds
                    
            abs_q_data = np.abs(q_data)
            bad_pixel_mask[:,masks[q_id]] = abs_q_data > (means + stds*max_sigma[0])[:,None]
            very_bad_pixel_mask[:,masks[q_id]] = abs_q_data > (means + stds*max_sigma[1])[:,None]
            
        for f_id in range(len(data)):
            very_bad_pixel_mask[f_id] = grow_mask(very_bad_pixel_mask[f_id],2)
        bad_pixel_mask |= very_bad_pixel_mask 
        data[bad_pixel_mask] = fill_value
        return data, bad_pixel_mask
    
    
    @staticmethod    
    def mask_bragg_pixels_old(data, masks, max_sigma, fill_value, use_log_scale,**kwargs):
        if use_log_scale:
            min_data = abs(data.min())+1
            #assumes min of data is 0
            q_sorted_data = tuple(np.log(min_data+data[:,q_mask]) for q_mask in masks)
        else:
            q_sorted_data = tuple(data[:,q_mask] for q_mask in masks)

        bad_pixel_mask = np.zeros(data.shape,dtype =bool)
        very_bad_pixel_mask = np.zeros(data.shape,dtype =bool)
        for q_id,q_data in enumerate(q_sorted_data):
            #log.info('q_data shape = {}'.format(q_data.shape))
            non_zero_mask = q_data != 0
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore')
                # gives divide by 0 error if q frame is compleatelly masked i.e. zero
                stds = np.std(q_data, axis = 1 , where = non_zero_mask)
                stds[np.isnan(stds)] = 0.0
                means = np.mean(q_data, axis = 1 , where = non_zero_mask)
                means[np.isnan(means)] = 0.0
            abs_q_data = np.abs(q_data)
            bad_pixel_mask[:,masks[q_id]] = abs_q_data > (means + stds*max_sigma[0])[:,None]
            very_bad_pixel_mask[:,masks[q_id]] = abs_q_data > (means + stds*max_sigma[1])[:,None]
            
        for f_id in range(len(data)):
            very_bad_pixel_mask[f_id] = grow_mask(very_bad_pixel_mask[f_id],2)
        bad_pixel_mask |= very_bad_pixel_mask 
        data[bad_pixel_mask] = fill_value
        return data, bad_pixel_mask
    
    @staticmethod    
    def count_bragg_pixels(data, masks, max_sigma,use_log_scale,**kwargs):
        #print((data != 0).any())
        if use_log_scale:
            min_data = abs(data.min())+1
            q_sorted_data = tuple(np.log(min_data+data[:,q_mask]) for q_mask in masks)
        else:
            q_sorted_data = tuple(data[:,q_mask] for q_mask in masks)
        stds = np.array(tuple( np.std(data_ring,axis = 1) for data_ring in q_sorted_data))
        #print(q_stored_data)
        # compare abs values in each q ring with sigma limit and count violations 
        bragg_counts_per_ring = np.array(tuple(np.sum(np.abs(data_ring)> (np.mean(data_ring, axis = 1)+std*max_sigma)[:,None],axis = 1) for data_ring,std in zip(q_sorted_data,stds)))
        bragg_counts = np.sum(bragg_counts_per_ring,axis = 0)
        #print(bragg_counts)
        return bragg_counts


class NormalizationFilter(Filter):
    def __init__(self,opt : dict):
        super().__init__(opt)
        
    def reset_data(self):
        self.data={}
        
    def _apply(self,calib_chunk,masks):
        #log.info('applying normalization')
        #log.info('masks = {}'.format(masks))
        roi_mask = self.roi_mask
        #good_frames = ~masks['total_filtered']
        #log.info('good_frames_shape = {}'.format(good_frames.shape))
        #log.info('any bad frames = {}'.format(good_frames.all()))
        data = calib_chunk['data']#[good_frames]
        #log.info('norm_data = {}'.format(data[0,0,0,:100]))
        #log.info('data_shape = {}'.format(data.shape))
        #log.info('data_chunk_shape = {}'.format(calib_chunk['data'].shape))
        #log.info('{} of points are masked by roi'.format(np.sum(~roi_mask)/np.prod(roi_mask.shape)))
        mask = calib_chunk['mask']#[good_frames]
        #mean_per_frame = masked_mean_new(data[:,roi_mask],mask[:,roi_mask],axis = 1)[0]
        #N=len(data)
        #mean_per_frame = masked_mean_new(data.reshape(N,-1),mask.reshape(N,-1),axis = 1)[0]
        #log.info('mean_per_frame = {}'.format(mean_per_frame))
        mean_per_frame = self.calc_mean(data,mask,roi_mask,axis = 1)
        scaling_factors = mean_per_frame
        calib_chunk['data'][:] /= scaling_factors[:,None,None,None]
        is_nan = np.isnan(scaling_factors)
        if is_nan.any():
            log.warning('Nan scalingfactors during application of NormalizationFilter')
            #log.info('scaling_factors shape {} first elements {}'.format(scaling_factors.shape,scaling_factors[:10]))

        out_mask = np.full(len(data),False)
        ## I cheet here a bit with out_mask. I don't count normalization as modification because otherwise all frames would always
        ## be viewd as modified if normalization is applied.
        ## In that sense normalization is not really a filter ... but its convenient to implent it as such if there are filters applied anyways.
        return calib_chunk,out_mask,out_mask

    def calc_mean(self,data,mask,roi_mask,axis = 1):
        #modules = convert_to_slice_if_possible(self.modules)
        # norm_data and norm_mask will be copies of the selected range not views becaus of numpy fancy indexing with booleans in roi_mask.
        #log.info('roi mask shape = {}'.format(roi_mask.shape))
        #log.info('mask shape = {}'.format(mask[:,modules].shape))
        #log.info('norm_data contains nan  = {} percentage = {}%'.format(np.isnan(data).any(),np.sum(np.isnan(data))/(np.prod(data.shape))*100 ))
        n_frames = len(data)
        norm_mask = (mask & roi_mask[None,...]).reshape(n_frames,-1)
        norm_data = data.reshape(n_frames,-1)
        #log.info('normalization roi is everything = {}'.format(self.normalization_mask.all()))
        #log.info('data mask containes {} % unmasked values'.format(np.sum(mask)/np.prod(mask.shape)))
        #log.info('norm mask containes {} % unmasked values'.format(np.sum(norm_mask)/np.prod(norm_mask.shape)))
        #log.info('mean ={}'.format(np.mean(norm_data,axis=axis)))
        means,counts = masked_mean(norm_data,norm_mask, axis = axis)
        #log.info(means)
        return np.atleast_1d(means)


class ADUFilter2D(Filter):
    def __init__(self,opt : DictNamespace):
        super().__init__(opt)
        self.limits = opt.limits
    def _apply(self,calib_chunk,masks):
        limits = self.limits
        data = calib_chunk['data']
        mask = calib_chunk['mask']        
        lower_mask = True
        upper_mask = True
        
        if isinstance(limits[0],(float,int)):
            lower_mask = data >= limits[0]
        if isinstance(limits[1],(float,int)):
            upper_mask = data <= limits[0]
        adu_mask = lower_mask & upper_mask
        new_mask = mask & adu_mask
        calib_chunk['mask'][:] = new_mask
        #log.info('negative data = {}'.format(data[~lower_mask].reshape(-1)[:100]))
        return calib_chunk,np.array(False),np.array(False)


class GainFilter2D(Filter):
    def __init__(self,opt : DictNamespace):
        super().__init__(opt)
        self.mask_opt = opt.mask
        self.gain_values= {'low':2,'medium':1,'high':0}
    def _apply(self,calib_chunk,masks):
        mask = calib_chunk['mask']
        gain = calib_chunk['gain']        

        gain_values = self.gain_values
        gain_masks = {'low':True,'medium':True,'high':True}

        gain_mask = False
        for key in enumerate(gain_values):
            if mask[key]:
                gain_mask = gain_mask | (gain == gain_valus[key])
                
        valid_mask = (~gain_mask)
        calib_chunk['mask'][:] = mask & valid_mask 
        return calib_chunk,False,False

    
class Filter1D(Filter):
    def __init__(self,opt : dict):
        super().__init__(opt)
        metrics = self.opt['metrics']
        if not isinstance(metrics,(tuple,list)):
            metrics = [metrics]
        self.metrics = metrics
        
        limits = self.opt['limits']
        if not isinstance(limits[0],(list,tuple)):
            limits = [limits]
        self.limits = limits
        
        
    def _apply(self,calib_chunk,masks):
        #good_frames = ~masks['total_filtered']
        data = calib_chunk['data']#[good_frames]
        mask = calib_chunk['mask']#[good_frames]
        metrics= self.metrics
        metric_values = self.calc_metric_values(data,mask)
        #log.info(metric_values)
        #log.info('max {} min {}'.format(metric_values[0].max(),metric_values[0].min()))
        #filtered_mask = np.zeros(len(calib_chunk['data']),dtype = bool)
        filtered_mask = self.apply_limits(metric_values[0])
        #log.info('{} of {} filtered'.format(np.sum(filtered_mask),len(filtered_mask)))
        return calib_chunk,filtered_mask,np.full(len(data),False)

    def calc_metric_values(self,data,mask,*args,**kwargs):
        metrics= self.metrics
        metric_values=np.zeros((len(data),len(metrics)))
        #log.info('data shape = {}'.format(data.shape))
        #log.info('mean data = {} '.format(np.mean(data, axis = tuple(range(1,len(data.shape))), where = mask)))
        for m_id in range(len(metrics)):
            metric=metrics[m_id]
            #metric_values[:,m_id] = np.mean(data, axis = tuple(range(1,len(data.shape))), where = mask)
            for f_id in range(len(data)):
                f_mask = mask[f_id]
                f_dat=data[f_id][f_mask]
                if len(f_dat) > 0:
                    metric_values[f_id,m_id] = metric(f_dat)
                else:
                    metric_values[f_id,m_id] = 0
        #log.info('metric values = {} '.format(metric_values))
        return  [metric_values]

    def apply_limits(self,metric_values):
        masks = np.zeros(metric_values.shape,dtype = bool)
        for index,limit in enumerate(self.limits):
            values = metric_values[:,index]
            if limit[0] == None:
                limit[0] = values.min()
            if limit[1] == None:
                limit[1] = values.max()
            #log.info('{} metric values are grater than {}'.format(np.sum(metric_values>limit[1]),limit[1]))
            #log.info('{} metric values are smaller than {}'.format(np.sum(metric_values<limit[0]),limit[0]))
            masks[:,index] = (values<limit[0]) | (values>limit[1])
        # logical and
        #log.info('number of unmasked values = {}'.format(np.sum(masks)))
        combined_mask = np.sum(np.array(masks), axis = 1).astype(bool)
        #log.info('number of unmasked values combined = {}'.format(np.sum(combined_mask)))
        return combined_mask


class LitPixels(Filter1D):
    def __init__(self,opt : dict):
        opt['metrics'] = [self.metric]
        super().__init__(opt)
        self.lit_threshold = self.opt['lit_threshold']

        
        limits = self.opt['limits']
        if not isinstance(limits[0],(list,tuple)):
            limits = [limits]
        self.limits = limits
    def metric(self,data,n_pixels,*args,**kwargs):
        lit_pixel_fraction = np.sum(data>self.lit_threshold)/n_pixels
        return lit_pixel_fraction
        
    
    def calc_metric_values(self,data,mask,*args,**kwargs):
        metrics= self.metrics
        metric_values=np.zeros((len(data),len(metrics)))
        #log.info('data shape = {}'.format(data.shape))
        #log.info('mean data = {} '.format(np.mean(data, axis = tuple(range(1,len(data.shape))), where = mask)))
        for m_id in range(len(metrics)):
            metric=metrics[m_id]
            #metric_values[:,m_id] = np.mean(data, axis = tuple(range(1,len(data.shape))), where = mask)
            for f_id in range(len(data)):
                f_mask = mask[f_id]
                n_pixels = np.prod(f_mask.shape)
                f_dat=data[f_id][f_mask]
                if len(f_dat) > 0:
                    metric_values[f_id,m_id] = metric(f_dat,n_pixels)
                else:
                    metric_values[f_id,m_id] = 0
        #log.info('metric values = {} '.format(metric_values))
        return  [metric_values]



import time
import numpy as np
from scipy.interpolate import griddata
import logging
import traceback

log=logging.getLogger('root')

from xframe.library.physicsLibrary import ewald_sphere_theta
from xframe.library.gridLibrary import GridFactory
from xframe.library.mathLibrary import cartesian_to_spherical
from xframe.library.mathLibrary import spherical_to_cartesian
from xframe.library import units
from xframe.library.gridLibrary import get_linspace_log2
from xframe.library.gridLibrary import uniformGrid_func

class AgipdRegridder:
    def __init__(self,db,geometry,wavelength, file_name='regrid_data', save_regrid_data=False, generate_data=False):
        #framed_pixel_grid in spherical_coordinates
        self.pixel_grid_spher = geometry['framed_pixel_grid']
        self.pixel_grid_spher[...,0]*=units.standardLength
        self.lab_pixel_grid_cart = geometry['framed_lab_pixel_grid']
        self.pixel_mask = geometry['framed_mask']
        self.data_shape = geometry['data_shape']
        self.wavelength = wavelength
        self.file_name = file_name
        self.db=db
        if generate_data:
            self.regrid_data=self.generate_regrid_data()
            if save_regrid_data:
                db.save(file_name,self.regrid_data)
        else:
            try:
                self.regrid_data=db.load(file_name)
            except FileNotFoundError as e:
                log.warning('regrid_data file not found. Proceed by generating regrid_data.')
                self.regrid_data=self.generate_regrid_data()
                if save_regrid_data:
                    db.save(file_name,self.regrid_data)

    def approximate_pixel_size(self,grid_cart):
        #absolute minimal x-distance between consecutive agipd pixels along vertical(short direction 128 pixels per module) pixel row
        x_min = np.abs([np.diff(grid_cart[m][:,:,0],axis=1).min() for m in np.arange(len(grid_cart))]).min()
        #same but in y-direction
        y_min = np.abs([np.diff(grid_cart[m][:,:,1],axis=1).min() for m in np.arange(len(grid_cart))]).min()
        #same but in z-direction
        z_min = np.abs([np.diff(grid_cart[m][:,:,2],axis=1).min() for m in np.arange(len(grid_cart))]).min()        
        approximated_pixel_size= np.max([x_min,y_min,z_min])
        log.info('approximate_pixel_size = {}'.format(approximated_pixel_size))
        return approximated_pixel_size


    def generate_regrid_data(self):
        pixel_grid_cart = spherical_to_cartesian(self.pixel_grid_spher)
        data = self.generate_regrid_data_polar(self.pixel_grid_spher,grid_point_multiplicator = [3,1])    
        data2 = self.generate_regrid_data_cart(self.lab_pixel_grid_cart,grid_point_multiplicator = [4,4])
        lab_small = self.generate_regrid_data_cart(self.lab_pixel_grid_cart,grid_point_multiplicator = [2,2])
        q_cart_small = self.generate_regrid_data_cart(pixel_grid_cart,grid_point_multiplicator = [2,2])
        
        cart_data = {'lab':data2,'lab_small':lab_small,'q_small':q_cart_small}
        polar_data = {'reciprocal':data}
        regrid_data = {'cart':cart_data,'polar':polar_data}
        return regrid_data


    def generate_regrid_data_cart(self,pixel_grid,grid_point_multiplicator=[4,4],on_q_space=False):
        regrid_data={}
        d_shape = self.data_shape
        
        pix_size=self.approximate_pixel_size(pixel_grid)        
        
        old_global_x_bounds = (pixel_grid[...,0].min(),pixel_grid[...,0].max())
        old_global_y_bounds = (pixel_grid[...,1].min(),pixel_grid[...,1].max())

        try:
            assert (old_global_x_bounds[0]<=0) and (old_global_x_bounds[1]>0), 'global x bounds: ( {} ) do not contain 0.'.format(old_global_x_bounds)
            assert (old_global_y_bounds[0]<=0) and (old_global_y_bounds[1]>0), 'global y bounds: ( {} ) do not contain 0.'.format(old_global_y_bounds)
        except AssertionError as e:
            log.error(e)
            traceback.print_exc()
            raise e

        target_x_step = pix_size/grid_point_multiplicator[0]
        xs,x_param = get_linspace_log2(old_global_x_bounds,target_x_step,additional_steps = 1,return_parameters = True)
        x_step = x_param['step_size']
        global_x_bounds = x_param['domain']

        target_y_step = pix_size/grid_point_multiplicator[1]
        ys,y_param = get_linspace_log2(old_global_y_bounds,target_y_step,additional_steps = 1, return_parameters = True)
        y_step = y_param['step_size']
        global_y_bounds = y_param['domain']

        out_grid = GridFactory.construct_grid('uniform',[xs,ys]).array
        
        regrid_data['x_step'] = x_step
        regrid_data['xs'] = xs
        regrid_data['y_step'] = y_step
        regrid_data['ys'] = ys
        regrid_data['grid'] = out_grid

        global_x_min = global_x_bounds[0]
        global_y_min = global_y_bounds[0]
        log.info('out grid shape = {}'.format(out_grid.shape))
        for m,g in enumerate(pixel_grid):
            log.info('generating regrid_data for module {}/{}'.format(m,len(pixel_grid)))
            regrid_data[str(m)]={}
            m_dict=regrid_data[str(m)]
        
            x_bounds = [g[...,0].min(),g[...,0].max()]            
            x_min_index = max(0,int((x_bounds[0]-global_x_min)//x_step)+1)
            x_max_index = min(int((x_bounds[1]-global_x_min)//x_step)+1,out_grid.shape[0])        

            y_bounds = [g[...,1].min(),g[...,1].max()]
            #log.info('x_bounds = {}'.format(x_bounds))
            y_min_index = max(int((y_bounds[0]-global_y_min)//y_step)+1,0)
            y_max_index = min(int((y_bounds[1]-global_y_min)//y_step)+1,out_grid.shape[1])        
            m_grid = out_grid[x_min_index:x_max_index,y_min_index:y_max_index]
            log.info("x_range = {} y_range ={}".format([x_max_index,x_min_index],[y_max_index,y_min_index]))
            #log.info("x_range = {} y_range ={}".format([x_max_index-x_min_index],[y_max_index-y_min_index]))
            log.info('mgrid.shape = {}'.format(m_grid.shape))
            old_m_grid =g[...,:2]
            old_index = np.zeros(g.shape[:-1])
            old_index[:-1,:-1][self.pixel_mask[m]] = np.arange(1,d_shape[1]*d_shape[2]+1)
        
            index = griddata(old_m_grid.reshape(-1,2),old_index.flatten(),m_grid.reshape(-1,2),method='nearest',rescale = True).reshape(m_grid.shape[:-1])
            #log.info('index.shape = {}'.format(index.shape))
            #log.info('ranges y: {} x:{}'.format(y_max_index-y_min_index,x_max_index - x_min_index))
            #new_data = griddata(old_grid,old_data.flatten(),new_grid,method='linear')

            m_dict['index']=index
            m_dict['x_range']=[x_min_index,x_max_index]
            m_dict['y_range']=[y_min_index,y_max_index]
            valid_index_mask = (index!=0)
            m_dict['valid_data_mask'] = valid_index_mask
            m_dict['valid_index'] = (index[valid_index_mask]-1).astype(int)            
        return regrid_data

    def generate_regrid_data_polar(self,pixel_grid,grid_point_multiplicator=[3,1],on_q_space=False):
        regrid_data={}
        d_shape = self.data_shape

        pixel_grid_c = spherical_to_cartesian(pixel_grid)
        pix_size=self.approximate_pixel_size(pixel_grid_c)
        
        old_global_r_bounds = (pixel_grid[...,0].min(),pixel_grid[...,0].max())
        old_global_phi_bounds = (pixel_grid[...,2].min(),pixel_grid[...,2].max())
        log.info('phi bounds = {}'.format(old_global_phi_bounds))

        target_r_step = pix_size/grid_point_multiplicator[0]
        r_length = old_global_r_bounds[1]-old_global_r_bounds[0]
        n_r_steps = int(r_length//target_r_step)
        r_step = r_length/(n_r_steps-1)
        rs = old_global_r_bounds[0]+ np.arange(n_r_steps)*r_step
        global_r_bounds = old_global_r_bounds
        log.info('r bounds = {} max,min = {}'.format(global_r_bounds,(rs.min(),rs.max())))


        # phi_step such that the arc length at max_x is proportional to the approximated pixel size    
        target_phi_step = pix_size/(global_r_bounds[1]*grid_point_multiplicator[1])
        phis,phi_param = get_linspace_log2(old_global_phi_bounds,target_phi_step,additional_steps = 1, return_parameters = True)
        phi_step = phi_param['step_size']
        global_phi_bounds = phi_param['domain']
        log.info('phi bounds = {} max,min = {}'.format(global_phi_bounds,(phis.min(),phis.max())))
                
        thetas = ewald_sphere_theta(self.wavelength,rs)
        tmp_grid = GridFactory.construct_grid('uniform',[rs,phis]).array
        out_grid = np.zeros(tmp_grid.shape[:-1]+(tmp_grid.shape[-1]+1,))
        out_grid[...,[0,2]] = tmp_grid
        out_grid[...,1] = thetas[:,None]
        
        
        regrid_data['r_step'] = r_step
        regrid_data['rs'] = rs
        regrid_data['phi_step'] = phi_step
        regrid_data['phis'] = phis
        regrid_data['grid'] = out_grid

        global_r_min = global_r_bounds[0]
        global_phi_min = global_phi_bounds[0]
        log.info('out grid shape = {}'.format(out_grid.shape))
        
        out_grid_c = spherical_to_cartesian(out_grid)[...,:2]
        pixel_grid_c = spherical_to_cartesian(pixel_grid)[...,:2]
        for m,g in enumerate(pixel_grid):            
            log.info('generating regrid_data for module {}/{}'.format(m,len(pixel_grid)))
            regrid_data[str(m)]={}
            m_dict=regrid_data[str(m)]

            g_c = pixel_grid_c[m]
        
            r_bounds = [g[...,0].min(),g[...,0].max()]            
            r_min_index = max(int((r_bounds[0]-global_r_min)//r_step)+1,0)
            r_max_index = min(int((r_bounds[1]-global_r_min)//r_step)+1,out_grid.shape[0])        

            phi_bounds = [g[...,2].min(),g[...,2].max()]
            #log.info('global phi_min = {}'.format(global_phi_min))
            #log.info('phi_bounds = {}'.format(phi_bounds))
            phi_min_index = max(int((phi_bounds[0]-global_phi_min)//phi_step)+1,0)
            phi_max_index = min(int((phi_bounds[1]-global_phi_min)//phi_step)+1,out_grid.shape[1])     
            out_g_c = out_grid_c[r_min_index:r_max_index,phi_min_index:phi_max_index]
            #log.info('min - glob =  {}'.format((phi_bounds[0]-global_phi_min)//phi_step+1))
            #log.info('max - glob =  {}'.format((phi_bounds[1]-global_phi_min)//phi_step+1))
            #log.info('out_grid shape = {}'.format(out_grid.shape))
            #log.info("r_range = {} phi_range ={}".format([r_max_index,r_min_index],[phi_max_index,phi_min_index]))
            #log.info("x_range = {} y_range ={}".format([x_max_index-x_min_index],[y_max_index-y_min_index]))
            log.info('mgrid.shape = {}'.format(out_g_c.shape))
            old_index = np.zeros(g.shape[:-1])
            old_index[:-1,:-1][self.pixel_mask[m]] = np.arange(1,d_shape[1]*d_shape[2]+1)
        
            index = griddata(g_c.reshape(-1,2),old_index.flatten(),out_g_c.reshape(-1,2),method='nearest',rescale = True).reshape(out_g_c.shape[:-1])
            #log.info('index nonzero: {}'.format((index!=0).any()))
            #log.info('index.shape = {}'.format(index.shape))
            #log.info('ranges y: {} x:{}'.format(y_max_index-y_min_index,x_max_index - x_min_index))
            #new_data = griddata(old_grid,old_data.flatten(),new_grid,method='linear')

            m_dict['index']=index
            m_dict['r_range']=[r_min_index,r_max_index]
            m_dict['phi_range']=[phi_min_index,phi_max_index]
            valid_index_mask = (index!=0)
            m_dict['valid_data_mask'] = valid_index_mask
            m_dict['valid_index'] = (index[valid_index_mask]-1).astype(int)            
        return regrid_data    
    
    def generate_regrid_data_polar_old(self,pixel_grid,grid_point_multiplicator=[2,1]):
        regrid_data={}
        d_shape = self.data_shape

        grid_c = spherical_to_cartesian(pixel_grid)
        pix_size=self.approximate_pixel_size(grid_c)
        
        r_max = pixel_grid[...,0].max()
        # r_step such that it is proportional to the approximated pixel size
        target_r_step = (pix_size/grid_point_multiplicator[0])
        n_new_rs = (r_max//target_r_step)
        r_step = r_max/n_new_rs
        rs= np.arange(n_new_rs)*r_step

        thetas = ewald_sphere_theta(self.wavelength,rs)

        # phi_step such that the arc length at max_x is proportional to the approximated pixel size
        
        target_phi_step = pix_size/(r_max*grid_point_multiplicator[1])
        phis, phi_param = get_linspace_log2([0,2*np.pi -target_phi_step],target_phi_step,additional_steps = 1,return_parameters = True)
        phi_step = phi_param['step_size']

        tmp_grid = GridFactory.construct_grid('uniform',[rs,phis]).array
        out_grid = np.zeros(tmp_grid.shape[:-1]+(tmp_grid.shape[-1]+1,))
        out_grid[...,[0,2]] = tmp_grid
        out_grid[...,1] = thetas[:,None]
        
        regrid_data['r_step'] = r_step
        regrid_data['rs'] = rs
        regrid_data['phi_step'] = phi_step
        regrid_data['phis'] = phis
        regrid_data['grid'] = out_grid
        log.info('out grid shape = {}'.format(out_grid.shape))
        log.info('out grid shape = {}'.format(out_grid.shape))
        grid_c = spherical_to_cartesian(out_grid)[...,:2]
        pixel_grid_c = spherical_to_cartesian(pixel_grid)[...,:2]
        
        for m,g in enumerate(pixel_grid):
            log.info('generating regrid_data for module {}/{}'.format(m,len(pixel_grid)))
            #log.info('g shape = {}'.format(g.shape))
            g_c=pixel_grid_c[m]
            #log.info('g_c shape = {}'.format(g_c.shape))
            regrid_data[str(m)]={}
            m_dict=regrid_data[str(m)]
        
            r_bounds = [g[...,0].min(),g[...,0].max()]
            r_min_index = int(r_bounds[0]//r_step)
            r_max_index = int(r_bounds[1]//r_step)        
            
            phi_bounds = [g[...,2].min(),g[...,2].max()]
            #log.info('phi_bounds = {}'.format(phi_bounds))
            phi_min_index = int((np.pi+phi_bounds[0])//phi_step)
            phi_max_index = int((np.pi+phi_bounds[1])//phi_step)        
            m_grid = grid_c[r_min_index:r_max_index,phi_min_index:phi_max_index]
            log.info("r_range = {} phi_range ={}".format([r_min_index,r_max_index],[phi_min_index,phi_max_index]))
            #log.info('mgrid.shape = {}'.format(m_grid.shape))
            old_m_grid =g_c
            old_index = np.zeros(g_c.shape[:-1])
            old_index[:-1,:-1][self.pixel_mask[m]] = np.arange(1,d_shape[1]*d_shape[2]+1)
            log.info('old index nonzero = {}'.format((old_index !=0).any()))
            #log.info('old m grid  = {}'.format(old_m_grid.shape))
            
            index = griddata(old_m_grid.reshape(-1,2),old_index.flatten(),m_grid.reshape(-1,2),method='nearest',rescale = True).reshape(m_grid.shape[:-1])
            log.info('index nonzero = {}'.format((index!=0).any()))
            #log.info('ranges r: {} phi:{}'.format(r_max_index-r_min_index,phi_max_index - phi_min_index))
            #log.info('index.shape = {}'.format(index.shape))
            #new_data = griddata(old_grid,old_data.flatten(),new_grid,method='linear')

            m_dict['index']=index
            m_dict['r_range']=[r_min_index,r_max_index]
            m_dict['phi_range']=[phi_min_index,phi_max_index]
            valid_index_mask = (index!=0)
            m_dict['valid_data_mask'] = valid_index_mask
            m_dict['valid_index'] = (index[valid_index_mask]-1).astype(int)            
        return regrid_data
    
    def generate_and_save_regrid_data(self):
        self.regrid_data=self.generate_regrid_data()
        self.db.save(self.file_name,self.regrid_data)

        

    def put_data_into_grid(self,data,out_grid,r_range,phi_range):
        new_data=np.zeros((data.shape[0],)+out_grid.shape[:-1])
        #log.info('new data shape = {}'.format(new_data.shape))
        #log.info("r_range = {} phi_range ={}".format(r_range,phi_range))
        #log.info("len r = {} len phi ={} data shape = {}".format(np.diff(r_range),np.diff(phi_range),data.shape))
        new_data[:,r_range[0]:r_range[1],phi_range[0]:phi_range[1]] = data
        return new_data
        
    def regrid_polar(self,data, combine_modules=False, modules_on_full_grid=False,grid_type='reciprocal'):
        regrid_polar_dict = self.regrid_data['polar']
        for known_grid_type in regrid_polar_dict:
            if grid_type == known_grid_type:
                regrid_data = regrid_polar_dict[known_grid_type]
                break
            
        out_grid = regrid_data['grid']
        out_grid_max_index = regrid_data['grid']
        
        out_data=[]
        for m in np.arange(len(data)):
            #log.info("m = {} of {}".format(m,len(data)))
            valid_data_mask=regrid_data[str(m)]['valid_data_mask'].astype(bool)
            #log.info('valid_data_mask shape = {}'.format(valid_data_mask.shape))
            valid_index=regrid_data[str(m)]['valid_index']
            #log.info('valid_index = {}'.format(valid_index.shape))

            if len(data.shape) == len(self.data_shape):
                data = data[None,...]
            regridded_m_data=np.zeros((data.shape[0],)+valid_data_mask.shape)
            regridded_m_data[:,valid_data_mask]=data[:,m].reshape(data.shape[0],-1)[:,valid_index]
            #log.info('valid_data_mask shape = {}'.format(valid_data_mask.shape))
            if (combine_modules or modules_on_full_grid):
                r_range = regrid_data[str(m)]['r_range']
                phi_range = regrid_data[str(m)]['phi_range']
                regridded_m_data = self.put_data_into_grid(regridded_m_data,out_grid,r_range,phi_range)[:,:-1,:-1]
            out_data.append(regridded_m_data)            
            
        if combine_modules:
            out_data = np.sum(out_data,axis=0)
        if modules_on_full_grid:
            out_data = np.array(out_data)
        if isinstance(out_data,np.ndarray):
            out_data = np.squeeze(out_data)
            
        return out_data
    
    def regrid_cart(self,data, combine_modules=False, modules_on_full_grid=False,grid_type='lab'):
        regrid_cart_dict = self.regrid_data['cart']
        #log.info(regrid_cart_dict.keys())
        for known_grid_type in regrid_cart_dict:
            if grid_type == known_grid_type:
                regrid_data = regrid_cart_dict[known_grid_type]
                break
        
        out_grid = regrid_data['grid']
        out_data=[]
        for m in np.arange(len(data)):
            #log.info("m = {} of {}".format(m,len(data)))
            valid_data_mask=regrid_data[str(m)]['valid_data_mask'].astype(bool)
            #log.info('valid_data_mask shape = {}'.format(valid_data_mask.shape))
            valid_index=regrid_data[str(m)]['valid_index']
            #log.info('valid_index = {}'.format(valid_index.shape))
            #log.info('valid_data_mask shape = {}'.format(valid_data_mask.shape))
            #log.info('regridded_m_data shape = {}'.format(regridded_m_data[~valid_data_mask].shape))
            if len(data.shape) == len(self.data_shape):
                data = data[None,...]
            regridded_m_data=np.zeros((data.shape[0],)+valid_data_mask.shape)
            regridded_m_data[:,valid_data_mask]=data[:,m].reshape(data.shape[0],-1)[:,valid_index]
            #log.info("regridded shape = {}".format(regridded_m_data.shape))
            
            if (combine_modules or modules_on_full_grid):
                r_range = regrid_data[str(m)]['x_range']
                phi_range = regrid_data[str(m)]['y_range']
                #log.info("x_range: {} y_range: {}".format(r_range,phi_range))
                regridded_m_data = self.put_data_into_grid(regridded_m_data,out_grid,r_range,phi_range)[:,:-1,:-1]                
                
            out_data.append(regridded_m_data)            
            
        if combine_modules:
            out_data = np.sum(out_data,axis=0)
        if modules_on_full_grid:
            out_data = np.array(out_data)
        if isinstance(out_data,np.ndarray):
            out_data = np.squeeze(out_data)
            
        return out_data



import numpy as np
import os
import glob
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.database.euxfel import default_DB
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import plot1D,heat2D_multi,heat2D,heatPolar2D,agipd_heat
from xframe.library.mathLibrary import circularHarmonicTransform_real_forward,cartesian_to_spherical,spherical_to_cartesian
from xframe.library.mathLibrary import circularHarmonicTransform_complex_forward

import logging
#from xframe.plugins.fxs3046_online.analysisLibrary.regrid2 import SimpleRegridder2D,AgipdRegridderSimple

from scipy.spatial.kdtree import KDTree
from scipy.ndimage import gaussian_filter

log=logging.getLogger('root')

init = False
if init:
    from xframe.startup_routines import dependency_injection
    dependency_injection()
    from xframe.startup_routines import load_recipes
    from xframe.control.Control import Controller
    analysis_sketch,exp_sketch = load_recipes(ana_name='fxs3046_online.generate_plots_2', ana_settings='saxs_65_-bg',exp_name='fxs3046_online',exp_settings = 'default_proc')
    controller = Controller(analysis_sketch,experiment_sketch = exp_sketch)
    controller.chooseJobSingleProcess()
    from xframe import database
    from xframe import Multiprocessing
    experiment = Multiprocessing.comm_module.get_experiment()

db = default_DB()

calib = experiment.calibrator
raw_base = '/gpfs/exfel/exp/SPB/202202/p003046/raw/'
runs = np.arange(160,170)
diod_path = 'INSTRUMENT/SPB_RR_SYS/ADC/UTC1-1:channel_0.output/data/rawData'
diod_train_path = 'INSTRUMENT/SPB_RR_SYS/ADC/UTC1-1:channel_0.output/data/trainId'
for run in runs:
    try:
        #calib.set_ids_for_run(run)
        raw_folder = raw_base + 'r{run:04d}/'.format(run = run)
        dataset_names = raw_folder + 'RAW-R{run:04d}-DIGI01-*.h5'.format(run = run)
        dset_paths = glob.glob(dataset_names)
        
        diod_list = []
        train_ids = []
        for path in dset_paths:
            print(path)
            with db.load(path,as_h5_object = True) as f:    
                d = f[diod_path][:]
                train_ids_part = f[diod_train_path][:]
                nonzero_mask = train_ids_part!=0
                max_value = np.max(d ,axis =1)
                diod_on = max_value>1000
                diod_list.append(diod_on[nonzero_mask])
                train_ids.append(train_ids_part[nonzero_mask])
        diod_array = np.concatenate(diod_list)    
        train_id_array = np.concatenate(train_ids)
        lookup = np.argsort(train_id_array)
        train_id_array = train_id_array[lookup]
        diod_array = diod_array[lookup]
        data = {'trainIds':train_id_array,'diod_on':diod_array}
        db.save('/gpfs/exfel/theory_group/user/berberic/p3046/pump_diod/r{run:04d}.h5'.format(run=run),data)
    except Exception as e:
        print(e)
    



import numpy as np
import time
from xframe.settings import general
general.n_control_worker=0
import xframe
xframe.setup_experiment('fxs3046_online','normalized_proc')
from psutil import virtual_memory

log = xframe.log
db = xframe.database.experiment
opt = {
    'run':169,
    'data_mode': 'proc',
    'frame_range': slice(None),
    'selection': {
        'cells': slice(1,177),
        'pulses': slice(None),
        'trains': slice(0,-100,1)
                  },
    'modules':np.arange(16),
    'n_frames':20000,
    'good_cells':np.arange(1,202),
    'in_multiples_of':False,
    'free_mem':virtual_memory().free,
    'n_processes':False
}

exp = xframe.experiment_worker
#exp.set_ids_for_run(169)

#c = db.create_vds_module(169,False,modules = np.arange(16),n_processes = 10)
#c = db.create_vds(169,False,n_processes = 10)

c = next(exp.get_data(opt))

#c = db.load('data_chunk',frame_ids = np.arange(1,10001), **opt)
#for c_id , c in enumerate(exp.get_data(opt)):
#    del(c)
#    log.info('chunk {} done!'.format(c_id))

#start = time.time()
#c = db.load_chunks('chunks',**opt)
#log.info(time.time()-start)
#for i in range(169,170):
#    db.create_vds(i,False,create_modulewise_vds = True, create_complete_vds = False,n_processes = 40)



if False:
    def lload(m,run,N,**kwargs):
        f = db.load('vds_module',skip_custom_methods = True ,as_h5_object=True,path_modifiers={'run':run,'data_mode':'proc','module':m})['INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/data']
        out = np.zeros((N,)+f.shape[1:],dtype = f.dtype)
        f.read_direct(out,slice(100,N+100))
        return 1
    def load(run,N,**kwargs):
        f = db.load('vds',skip_custom_methods = True ,as_h5_object=True,path_modifiers={'run':run,'data_mode':'proc'})['INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/data']
        out = np.zeros((16,N)+f.shape[2:],dtype = f.dtype)
        f.read_direct(out,(slice(None),slice(100,N+100)))
        return 1
    
    runs = np.array([112,113,114,115])
    runs = np.arange(4000)
    N = 1000
    start = time.time()
    out = np.zeros((16,N)+f.shape[2:],dtype = f.dtype)
    #load(169,N)
    xframe.Multiprocessing.comm_module.request_mp_evaluation(lload,argArrays=[np.arange(16)],const_args=[169,N],n_processes=16,callWithMultipleArguments=False,assemble_outputs = False)
    stop = time.time()-start
    print(stop)
